import copy
import logging
from datetime import datetime
from dateutil.relativedelta import relativedelta

from spaceone.core import cache, utils, queue
from spaceone.core.manager import BaseManager
from spaceone.inventory.error.cloud_service_query_set import *
from spaceone.inventory.model.cloud_service_query_set_model import CloudServiceQuerySet
from spaceone.inventory.manager.cloud_service_manager import CloudServiceManager
from spaceone.inventory.manager.cloud_service_stats_manager import CloudServiceStatsManager

_LOGGER = logging.getLogger(__name__)

_DEFAULT_GROUP_BY = ['provider', 'cloud_service_group', 'cloud_service_type', 'region_code', 'account', 'project_id']


class CloudServiceQuerySetManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cloud_svc_query_set_model: CloudServiceQuerySet = self.locator.get_model('CloudServiceQuerySet')
        self.cloud_svc_stats_mgr = None

    @staticmethod
    def push_task(domain_id, system_token):
        task = {
            'name': 'run_query_sets_by_domain',
            'version': 'v1',
            'executionEngine': 'BaseWorker',
            'stages': [{
                'locator': 'SERVICE',
                'name': 'CloudServiceQuerySetService',
                'metadata': {
                    'service': 'inventory',
                    'resource': 'CloudServiceQuerySet',
                    'verb': 'run_query_sets_by_domain',
                    'token': system_token
                },
                'method': 'run_query_sets_by_domain',
                'params': {
                    'params': {
                        'domain_id': domain_id
                    }
                }
            }]
        }

        _LOGGER.debug(f'[push_task] run query sets by domain: {domain_id}')

        queue.put('collector_q', utils.dump_json(task))

    def create_cloud_service_query_set(self, params):
        def _rollback(cloud_svc_query_set_vo: CloudServiceQuerySet):
            _LOGGER.info(
                f'[ROLLBACK] Delete Cloud Service Query Set : {cloud_svc_query_set_vo.name} ({cloud_svc_query_set_vo.query_set_id})')
            cloud_svc_query_set_vo.delete()

        params['query_hash'] = utils.dict_to_hash(params['query_options'])

        _LOGGER.debug(f'[create_cloud_service_query_set] create query set: {params["name"]}')

        keys, additional_info_keys = self._get_keys_from_query(params['query_options'])
        params['keys'] = keys
        params['additional_info_keys'] = additional_info_keys

        provider = params.get('provider')
        cloud_service_group = params.get('cloud_service_group')
        cloud_service_type = params.get('cloud_service_type')

        if provider and cloud_service_group and cloud_service_type:
            params['ref_cloud_service_type'] = self._make_cloud_service_type_key(params['domain_id'], provider,
                                                                                 cloud_service_group,
                                                                                 cloud_service_type)

        cloud_svc_query_set_vo: CloudServiceQuerySet = self.cloud_svc_query_set_model.create(params)
        self.transaction.add_rollback(_rollback, cloud_svc_query_set_vo)

        # Check Analyze Query
        self._run_analyze_query(cloud_svc_query_set_vo)

        return cloud_svc_query_set_vo

    def update_cloud_service_query_set(self, params):
        return self.update_cloud_service_query_set_by_vo(
            params, self.get_cloud_service_query_set(params['query_set_id'], params['domain_id']))

    def update_cloud_service_query_set_by_vo(self, params, cloud_svc_query_set_vo: CloudServiceQuerySet):
        def _rollback(old_data):
            _LOGGER.info(f'[ROLLBACK] Revert Data : {old_data.get("query_set_id")}')
            cloud_svc_query_set_vo.update(old_data)

        if 'query_options' in params:
            params['query_hash'] = utils.dict_to_hash(params['query_options'])
            keys, additional_info_keys = self._get_keys_from_query(params['query_options'])
            params['keys'] = keys
            params['additional_info_keys'] = additional_info_keys

        _LOGGER.debug(f'[update_cloud_service_query_set_by_vo] update query set: {cloud_svc_query_set_vo.query_set_id}')

        self.transaction.add_rollback(_rollback, cloud_svc_query_set_vo.to_dict())

        if 'query_options' in params:
            # Check Analyze Query
            self._run_analyze_query(cloud_svc_query_set_vo)

        return cloud_svc_query_set_vo.update(params)

    def delete_cloud_service_query_set(self, query_set_id, domain_id):
        self.delete_cloud_service_query_set_by_vo(self.get_cloud_service_query_set(query_set_id, domain_id))

    def delete_cloud_service_query_set_by_vo(self, cloud_svc_query_set_vo: CloudServiceQuerySet):
        cloud_svc_stats_mgr: CloudServiceStatsManager = self.locator.get_manager('CloudServiceStatsManager')

        query_set_id = cloud_svc_query_set_vo.query_set_id
        domain_id = cloud_svc_query_set_vo.domain_id

        _LOGGER.debug(f'[delete_cloud_service_query_set_by_vo] delete query set: {query_set_id}')

        # Delete Cloud Service Stats Data
        stats_vos = cloud_svc_stats_mgr.filter_cloud_service_stats(query_set_id=query_set_id, domain_id=domain_id)
        stats_vos.delete()

        # Delete Monthly Cloud Service Stats Data
        monthly_stats_vos = cloud_svc_stats_mgr.filter_monthly_cloud_service_stats(query_set_id=query_set_id,
                                                                                   domain_id=domain_id)
        monthly_stats_vos.delete()

        cloud_svc_query_set_vo.delete()

    def get_cloud_service_query_set(self, query_set_id, domain_id, only=None):
        return self.cloud_svc_query_set_model.get(query_set_id=query_set_id, domain_id=domain_id, only=only)

    def filter_cloud_service_query_sets(self, **conditions):
        return self.cloud_svc_query_set_model.filter(**conditions)

    def list_cloud_service_query_sets(self, query):
        return self.cloud_svc_query_set_model.query(**query)

    def stat_cloud_service_query_sets(self, query):
        return self.cloud_svc_query_set_model.stat(**query)

    def run_cloud_service_query_set(self, cloud_svc_query_set_vo: CloudServiceQuerySet):
        if cloud_svc_query_set_vo.state == 'DISABLED':
            raise ERROR_CLOUD_SERVICE_QUERY_SET_STATE(state=cloud_svc_query_set_vo.state)

        self.cloud_svc_stats_mgr: CloudServiceStatsManager = self.locator.get_manager('CloudServiceStatsManager')

        _LOGGER.debug(f'[run_cloud_service_query_set] run query set: {cloud_svc_query_set_vo.query_set_id} '
                      f'({cloud_svc_query_set_vo.domain_id})')
        results = self._run_analyze_query(cloud_svc_query_set_vo)

        created_at = datetime.utcnow()

        try:
            self._save_query_results(cloud_svc_query_set_vo, results, created_at)
            self._delete_changed_cloud_service_stats(cloud_svc_query_set_vo, created_at)
            self._delete_changed_monthly_cloud_service_stats(cloud_svc_query_set_vo, created_at)
        except Exception as e:
            _LOGGER.error(f'[run_cloud_service_query_set] Failed to save query result: {e}', exc_info=True)
            self._rollback_query_results(cloud_svc_query_set_vo, created_at)
            raise ERROR_CLOUD_SERVICE_QUERY_SET_RUN_FAILED(query_set_id=cloud_svc_query_set_vo.query_set_id)

        self._update_status(cloud_svc_query_set_vo, created_at)
        self._delete_invalid_cloud_service_stats(cloud_svc_query_set_vo)
        self._delete_old_cloud_service_stats(cloud_svc_query_set_vo)
        self._remove_analyze_cache(cloud_svc_query_set_vo.domain_id, cloud_svc_query_set_vo.query_set_id)

    def test_cloud_service_query_set(self, cloud_svc_query_set_vo: CloudServiceQuerySet):
        if cloud_svc_query_set_vo.state == 'DISABLED':
            raise ERROR_CLOUD_SERVICE_QUERY_SET_STATE(state=cloud_svc_query_set_vo.state)

        self.cloud_svc_stats_mgr: CloudServiceStatsManager = self.locator.get_manager('CloudServiceStatsManager')

        _LOGGER.debug(f'[test_cloud_service_query_set] test query set: {cloud_svc_query_set_vo.query_set_id} '
                      f'({cloud_svc_query_set_vo.domain_id})')
        return {
            'results': self._run_analyze_query(cloud_svc_query_set_vo)
        }

    def _run_analyze_query(self, cloud_svc_query_set_vo: CloudServiceQuerySet):
        cloud_svc_mgr: CloudServiceManager = self.locator.get_manager('CloudServiceManager')

        analyze_query = copy.deepcopy(cloud_svc_query_set_vo.query_options)
        provider = cloud_svc_query_set_vo.provider
        cloud_service_group = cloud_svc_query_set_vo.cloud_service_group
        cloud_service_type = cloud_svc_query_set_vo.cloud_service_type

        analyze_query['filter'] = analyze_query.get('filter', [])
        analyze_query['filter'] += self._make_query_filter(provider, cloud_service_group, cloud_service_type)

        analyze_query['group_by'] = analyze_query.get('group_by', []) + _DEFAULT_GROUP_BY

        if 'select' in analyze_query:
            for group_by_key in _DEFAULT_GROUP_BY:
                analyze_query['select'][group_by_key] = group_by_key

        _LOGGER.debug(f'[run_cloud_service_query_set] Run Analyze Query: {analyze_query}')
        response = cloud_svc_mgr.analyze_cloud_services(analyze_query)
        return response.get('results', [])

    def _delete_invalid_cloud_service_stats(self, cloud_svc_query_set_vo: CloudServiceQuerySet):
        domain_id = cloud_svc_query_set_vo.domain_id
        query_set_id = cloud_svc_query_set_vo.query_set_id

        cloud_stats_vos = self.cloud_svc_stats_mgr.filter_cloud_service_stats(
            query_set_id=query_set_id, domain_id=domain_id, status='IN_PROGRESS')

        if cloud_stats_vos.count() > 0:
            _LOGGER.debug(f'[_delete_invalid_cloud_service_stats] delete stats count: {cloud_stats_vos.count()}')
            cloud_stats_vos.delete()

        monthly_stats_vos = self.cloud_svc_stats_mgr.filter_monthly_cloud_service_stats(
            query_set_id=query_set_id, domain_id=domain_id, status='IN_PROGRESS')

        if monthly_stats_vos.count() > 0:
            _LOGGER.debug(f'[_delete_invalid_cloud_service_stats] delete monthly stats count: {monthly_stats_vos.count()}')
            monthly_stats_vos.delete()

    def _delete_old_cloud_service_stats(self, cloud_svc_query_set_vo: CloudServiceQuerySet):
        now = datetime.utcnow().date()
        query_set_id = cloud_svc_query_set_vo.query_set_id
        domain_id = cloud_svc_query_set_vo.domain_id
        old_created_month = (now - relativedelta(months=12)).strftime('%Y-%m')
        old_created_year = (now - relativedelta(months=36)).strftime('%Y')

        delete_query = {
            'filter': [
                {'k': 'created_month', 'v': old_created_month, 'o': 'lt'},
                {'k': 'query_set_id', 'v': query_set_id, 'o': 'eq'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'}
            ]
        }

        stats_vos, total_count = self.cloud_svc_stats_mgr.list_cloud_service_stats(delete_query)

        if total_count > 0:
            _LOGGER.debug(f'[delete_old_cloud_service_stats] delete stats count: {total_count}')
            stats_vos.delete()

        monthly_delete_query = {
            'filter': [
                {'k': 'created_year', 'v': old_created_year, 'o': 'lt'},
                {'k': 'query_set_id', 'v': query_set_id, 'o': 'eq'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'}
            ]
        }

        monthly_stats_vos, total_count = self.cloud_svc_stats_mgr.list_monthly_cloud_service_stats(monthly_delete_query)

        if total_count > 0:
            _LOGGER.debug(f'[_delete_old_cloud_service_stats] delete monthly stats count: {total_count}')
            monthly_stats_vos.delete()

    def _update_status(self, cloud_svc_query_set_vo: CloudServiceQuerySet, created_at):
        domain_id = cloud_svc_query_set_vo.domain_id
        query_set_id = cloud_svc_query_set_vo.query_set_id
        created_date = created_at.strftime('%Y-%m-%d')
        created_month = created_at.strftime('%Y-%m')

        cloud_stats_vos = self.cloud_svc_stats_mgr.filter_cloud_service_stats(
            query_set_id=query_set_id, domain_id=domain_id, created_date=created_date, status='IN_PROGRESS')
        cloud_stats_vos.update({'status': 'DONE'})

        monthly_stats_vos = self.cloud_svc_stats_mgr.filter_monthly_cloud_service_stats(
            query_set_id=query_set_id, domain_id=domain_id, created_month=created_month, status='IN_PROGRESS')
        monthly_stats_vos.update({'status': 'DONE'})

    def _delete_changed_cloud_service_stats(self, cloud_svc_query_set_vo: CloudServiceQuerySet, created_at):
        domain_id = cloud_svc_query_set_vo.domain_id
        query_set_id = cloud_svc_query_set_vo.query_set_id
        created_date = created_at.strftime('%Y-%m-%d')

        cloud_stats_vos = self.cloud_svc_stats_mgr.filter_cloud_service_stats(
            query_set_id=query_set_id, domain_id=domain_id, created_date=created_date, status='DONE')

        _LOGGER.debug(f'[_delete_old_cloud_service_stats] delete count: {cloud_stats_vos.count()}')
        cloud_stats_vos.delete()

    def _delete_changed_monthly_cloud_service_stats(self, cloud_svc_query_set_vo: CloudServiceQuerySet, created_at):
        domain_id = cloud_svc_query_set_vo.domain_id
        query_set_id = cloud_svc_query_set_vo.query_set_id
        created_month = created_at.strftime('%Y-%m')

        monthly_stats_vos = self.cloud_svc_stats_mgr.filter_monthly_cloud_service_stats(
            query_set_id=query_set_id, domain_id=domain_id, created_month=created_month, status='DONE')

        _LOGGER.debug(f'[_delete_old_monthly_cloud_service_stats] delete count: {monthly_stats_vos.count()}')
        monthly_stats_vos.delete()

    def _rollback_query_results(self, cloud_svc_query_set_vo: CloudServiceQuerySet, created_at):
        _LOGGER.debug(f'[_rollback_query_results] Rollback Query Results: {cloud_svc_query_set_vo.query_set_id}')
        query_set_id = cloud_svc_query_set_vo.query_set_id
        domain_id = cloud_svc_query_set_vo.domain_id

        cloud_service_stats_vo = self.cloud_svc_stats_mgr.filter_cloud_service_stats(
            query_set_id=query_set_id, domain_id=domain_id, created_date=created_at.strftime('%Y-%m-%d'),
            status='IN_PROGRESS')
        cloud_service_stats_vo.delete()

        monthly_stats_vo = self.cloud_svc_stats_mgr.filter_monthly_cloud_service_stats(
            query_set_id=query_set_id, domain_id=domain_id, created_month=created_at.strftime('%Y-%m'),
            status='IN_PROGRESS')
        monthly_stats_vo.delete()

    def _save_query_results(self, query_set_vo: CloudServiceQuerySet, results, created_at):
        for result in results:
            self._save_query_result(result, query_set_vo, created_at)

    def _save_query_result(self, result, query_set_vo: CloudServiceQuerySet, created_at):
        provider = result['provider']
        cloud_service_group = result['cloud_service_group']
        cloud_service_type = result['cloud_service_type']
        region_code = result.get('region_code')
        query_set_id = query_set_vo.query_set_id
        domain_id = query_set_vo.domain_id
        ref_cloud_service_type = self._make_cloud_service_type_key(domain_id, provider, cloud_service_group,
                                                                   cloud_service_type)
        ref_region = self._make_region_key(domain_id, provider, region_code)

        data = {
            'query_set_id': query_set_id,
            'values': {},
            'unit': {},
            'provider': provider,
            'cloud_service_group': cloud_service_group,
            'cloud_service_type': cloud_service_type,
            'ref_cloud_service_type': ref_cloud_service_type,
            'region_code': region_code,
            'ref_region': ref_region,
            'account': result.get('account'),
            'project_id': result.get('project_id'),
            'domain_id': domain_id,
            'additional_info': {},
            'created_at': created_at,
            'created_year': created_at.strftime('%Y'),
            'created_month': created_at.strftime('%Y-%m'),
            'created_date': created_at.strftime('%Y-%m-%d')
        }

        for key in query_set_vo.keys:
            data['values'][key] = result.get(key, 0)

            if key in query_set_vo.unit:
                data['unit'][key] = query_set_vo.unit[key]
            else:
                data['unit'][key] = 'Count'

        for key in query_set_vo.additional_info_keys:
            data['additional_info'][key] = result.get(key)

        self.cloud_svc_stats_mgr.create_cloud_service_stats(data, False)
        self.cloud_svc_stats_mgr.create_monthly_cloud_service_stats(data, False)

    @staticmethod
    def _remove_analyze_cache(domain_id, query_set_id):
        cache.delete_pattern(f'inventory:cloud-service-stats:*:{domain_id}:{query_set_id}:*')
        cache.delete_pattern(f'inventory:stats-query-history:{domain_id}:{query_set_id}:*')

    @staticmethod
    def _make_cloud_service_type_key(domain_id, provider, cloud_service_group, cloud_service_type):
        return f'{domain_id}.{provider}.{cloud_service_group}.{cloud_service_type}'

    @staticmethod
    def _make_region_key(domain_id, provider, region_code):
        if region_code:
            return f'{domain_id}.{provider}.{region_code}'
        else:
            return None

    @staticmethod
    def _make_query_filter(provider=None, cloud_service_group=None, cloud_service_type=None):
        _filter = []

        if provider:
            _filter.append({
                'k': 'provider',
                'v': provider,
                'o': 'eq'
            })

        if cloud_service_group:
            _filter.append({
                'k': 'cloud_service_group',
                'v': cloud_service_group,
                'o': 'eq'
            })

        if cloud_service_type:
            _filter.append({
                'k': 'cloud_service_type',
                'v': cloud_service_type,
                'o': 'eq'
            })

        return _filter

    @staticmethod
    def _get_keys_from_query(query):
        keys = query.get('fields', {}).keys()
        additional_info_keys = []
        for key in query.get('group_by', []):
            if key not in _DEFAULT_GROUP_BY:
                additional_info_keys.append(key.split('.')[-1:][0])
        return keys, additional_info_keys



from spaceone.api.inventory.v1 import cloud_service_stats_pb2, cloud_service_stats_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class CloudServiceStats(BaseAPI, cloud_service_stats_pb2_grpc.CloudServiceStatsServicer):

    pb2 = cloud_service_stats_pb2
    pb2_grpc = cloud_service_stats_pb2_grpc

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceStatsService', metadata) as cloud_svc_stats_service:
            cloud_svc_type_vos, total_count = cloud_svc_stats_service.list(params)
            return self.locator.get_info('CloudServiceStatsInfo',
                                         cloud_svc_type_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def analyze(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceStatsService', metadata) as cloud_svc_stats_service:
            return self.locator.get_info('AnalyzeInfo', cloud_svc_stats_service.analyze(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceStatsService', metadata) as cloud_svc_stats_service:
            return self.locator.get_info('StatisticsInfo', cloud_svc_stats_service.stat(params))



import functools
from spaceone.api.inventory.v1 import cloud_service_query_set_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.inventory.model.cloud_service_query_set_model import CloudServiceQuerySet

__all__ = ['CloudServiceQuerySetInfo', 'CloudServiceQuerySetsInfo']


def CloudServiceQuerySetInfo(cloud_svc_query_set_vo: CloudServiceQuerySet, minimal=False):
    info = {
        'query_set_id': cloud_svc_query_set_vo.query_set_id,
        'name': cloud_svc_query_set_vo.name,
        'state': cloud_svc_query_set_vo.state,
        'query_type': cloud_svc_query_set_vo.query_type,
        'provider': cloud_svc_query_set_vo.provider,
        'cloud_service_group': cloud_svc_query_set_vo.cloud_service_group,
        'cloud_service_type': cloud_svc_query_set_vo.cloud_service_type,
    }

    if not minimal:
        info.update({
            'query_options': change_analyze_query(cloud_svc_query_set_vo.query_options),
            'keys': cloud_svc_query_set_vo.keys,
            'additional_info_keys': cloud_svc_query_set_vo.additional_info_keys,
            'unit': change_struct_type(cloud_svc_query_set_vo.unit),
            'tags': change_struct_type(cloud_svc_query_set_vo.tags),
            'domain_id': cloud_svc_query_set_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(cloud_svc_query_set_vo.created_at),
            'updated_at': utils.datetime_to_iso8601(cloud_svc_query_set_vo.updated_at)
        })

    return cloud_service_query_set_pb2.CloudServiceQuerySetInfo(**info)


def CloudServiceQuerySetsInfo(cloud_svc_query_set_vos, total_count, **kwargs):
    return cloud_service_query_set_pb2.CloudServiceQuerySetsInfo(
        results=list(map(functools.partial(CloudServiceQuerySetInfo, **kwargs), cloud_svc_query_set_vos)),
        total_count=total_count)



import logging

from spaceone.core.service import *
from spaceone.inventory.model.record_model import Record
from spaceone.inventory.model.cloud_service_model import CloudService
from spaceone.inventory.manager.record_manager import RecordManager
from spaceone.inventory.manager.note_manager import NoteManager
from spaceone.inventory.manager.cloud_service_manager import CloudServiceManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class NoteService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.note_mgr: NoteManager = self.locator.get_manager('NoteManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['record_id', 'note', 'domain_id'])
    def create(self, params):
        """Create record note

        Args:
            params (dict): {
                'record_id': 'str',
                'note': 'str',
                'domain_id': 'str'
            }

        Returns:
            note_vo (object)
        """

        user_id = self.transaction.get_meta('user_id')

        cloud_svc_mgr: CloudServiceManager = self.locator.get_manager('CloudServiceManager')
        record_mgr: RecordManager = self.locator.get_manager('RecordManager')
        record_vo: Record = record_mgr.get_record(params['record_id'], params['domain_id'])
        cloud_svc_vo: CloudService = cloud_svc_mgr.get_cloud_service(record_vo.cloud_service_id, params['domain_id'])

        params['cloud_service_id'] = record_vo.cloud_service_id
        params['project_id'] = cloud_svc_vo.project_id
        params['created_by'] = user_id

        return self.note_mgr.create_note(params)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['note_id', 'domain_id'])
    def update(self, params):
        """Update record note

        Args:
            params (dict): {
                'note_id': 'str',
                'note': 'dict',
                'domain_id': 'str'
            }

        Returns:
            note_vo (object)
        """
        
        note_id = params['note_id']
        domain_id = params['domain_id']

        note_vo = self.note_mgr.get_note(note_id, domain_id)

        # Check permission

        return self.note_mgr.update_note_by_vo(params, note_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['note_id', 'domain_id'])
    def delete(self, params):
        """Delete record note

        Args:
            params (dict): {
                'note_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.note_mgr.delete_note(params['note_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['note_id', 'domain_id'])
    def get(self, params):
        """ Get record note

        Args:
            params (dict): {
                'note_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            note_vo (object)
        """

        return self.note_mgr.get_note(params['note_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['domain_id'])
    @append_query_filter(['note_id', 'record_id', 'cloud_service_id', 'created_by', 'domain_id', 'user_projects'])
    @append_keyword_filter(['note'])
    def list(self, params):
        """ List record notes

        Args:
            params (dict): {
                'note_id': 'str',
                'record_id': 'str',
                'cloud_service_id': 'str',
                'created_by': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)',
                'user_projects': 'list', // from meta
            }

        Returns:
            note_vos (object)
            total_count
        """

        query = params.get('query', {})
        return self.note_mgr.list_notes(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_projects'])
    @append_keyword_filter(['note'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.note_mgr.stat_notes(query)



import logging

from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.inventory.connector.config_connector import ConfigConnector

_LOGGER = logging.getLogger(__name__)


class ConfigManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_conn: SpaceConnector = self.locator.get_connector('SpaceConnector', service='config')

    def get_domain_config(self, name, domain_id):
        result = self.config_conn.dispatch('DomainConfig.get', {'name': name, 'domain_id': domain_id})

        rules = result['data']['rules']
        policies = {}
        for rule in rules:
            policies[rule['k']] = int(rule['v'])
        _LOGGER.debug(f'[get_domain_config] policies : {policies}')
        return policies



import logging
from spaceone.core.manager import BaseManager
from spaceone.inventory.model.collection_state_model import CollectionState

_LOGGER = logging.getLogger(__name__)


class CollectionStateManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.collector_id = self.transaction.get_meta('collector_id')
        self.job_task_id = self.transaction.get_meta('job_task_id')
        self.secret_id = self.transaction.get_meta('secret.secret_id')
        self.collection_state_model: CollectionState = self.locator.get_model('CollectionState')

    def create_collection_state(self, resource_id, resource_type, domain_id):
        def _rollback(state_vo: CollectionState):
            _LOGGER.info(f'[ROLLBACK] Delete Collection State : resource_id = {state_vo.resource_id}, '
                         f'collector_id = {state_vo.collector_id}')
            state_vo.terminate()

        if self.collector_id and self.job_task_id and self.secret_id:
            state_data = {
                'collector_id': self.collector_id,
                'job_task_id': self.job_task_id,
                'secret_id': self.secret_id,
                'resource_id': resource_id,
                'resource_type': resource_type,
                'domain_id': domain_id
            }

            # _LOGGER.debug(f'[create_collection_state] create collection state: {state_data}')
            state_vo = self.collection_state_model.create(state_data)
            self.transaction.add_rollback(_rollback, state_vo)

    def update_collection_state_by_vo(self, params, state_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[ROLLBACK] Revert Collection State : resource_id={state_vo.resource_id}, '
                         f'collector_id={state_vo.collector_id}')
            state_vo.update(old_data)

        self.transaction.add_rollback(_rollback, state_vo.to_dict())
        return state_vo.update(params)

    def reset_collection_state(self, state_vo):
        if self.job_task_id:
            params = {
                'disconnected_count': 0,
                'job_task_id': self.job_task_id
            }

            self.update_collection_state_by_vo(params, state_vo)

    def get_collection_state(self, resource_id, domain_id):
        if self.collector_id and self.secret_id:
            state_vos = self.collection_state_model.filter(collector_id=self.collector_id,
                                                           secret_id=self.secret_id,
                                                           resource_id=resource_id,
                                                           domain_id=domain_id)

            if state_vos.count() > 0:
                return state_vos[0]

        return None

    def filter_collection_states(self, **conditions):
        return self.collection_state_model.filter(**conditions)

    def list_collection_states(self, query):
        return self.collection_state_model.query(**query)

    def delete_collection_state_by_resource_id(self, resource_id, domain_id):
        # _LOGGER.debug(f'[delete_collection_state_by_resource_id] delete collection state: {resource_id}')
        state_vos = self.collection_state_model.filter(resource_id=resource_id, domain_id=domain_id)
        state_vos.delete()

    def delete_collection_state_by_resource_ids(self, resource_ids):
        # _LOGGER.debug(f'[delete_collection_state_by_resource_ids] delete collection state: {resource_ids}')
        _filter = [{'k': 'resource_id', 'v': resource_ids, 'o': 'in'}]
        state_vos, total_count = self.collection_state_model.query(filter=_filter)
        state_vos.delete()

    def delete_collection_state_by_collector_id(self, collector_id, domain_id):
        # _LOGGER.debug(f'[delete_collection_state_by_collector_id] delete collection state: {collector_id}')
        state_vos = self.collection_state_model.filter(collector_id=collector_id, domain_id=domain_id)
        state_vos.delete()



import ipaddress


class IPAddress(object):

    @staticmethod
    def get_ip_object(ip_address):
        return ipaddress.ip_address(ip_address)

    @staticmethod
    def get_network_object(network):
        return ipaddress.ip_network(network)

    @staticmethod
    def check_valid_ip(ip_address):
        try:
            ip = ipaddress.ip_address(ip_address)
            return str(ip)
        except ValueError:
            return False

    @staticmethod
    def check_valid_network(cidr):
        try:
            network = ipaddress.ip_network(cidr)
            return str(network)
        except ValueError:
            return False

    @staticmethod
    def check_duplicate_cidr_range(cidr1, cidr2):
        """
        if CIDRs was duplicated, return True
        """
        try:
            c1 = ipaddress.ip_network(cidr1)
            c2 = ipaddress.ip_network(cidr2)

            if c1.subnet_of(c2) or c1.supernet_of(c2):
                return True

            return False

        except ValueError:
            return False

    @staticmethod
    def check_subnet_of_network(subnet_cidr, net_cidr):
        subnet_cidr = ipaddress.ip_network(subnet_cidr)
        net_cidr = ipaddress.ip_network(net_cidr)

        return subnet_cidr.subnet_of(net_cidr)

    @staticmethod
    def check_valid_ip_in_network(ip, cidr):
        _ip = ipaddress.ip_address(ip)
        _net = ipaddress.ip_network(cidr)

        if _ip in _net:
            return True

        return False



"""
Deprecated:
  Not used. Integrated with SpaceConnector.
"""

import logging

from google.protobuf.json_format import MessageToDict

from spaceone.core.connector import BaseConnector
from spaceone.core import pygrpc
from spaceone.core.utils import parse_grpc_endpoint
from spaceone.core.error import *


__all__ = ['IdentityConnector']
_LOGGER = logging.getLogger(__name__)


class IdentityConnector(BaseConnector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'endpoint' not in self.config:
            raise ERROR_WRONG_CONFIGURATION(key='endpoint')

        if len(self.config['endpoint']) > 1:
            raise ERROR_WRONG_CONFIGURATION(key='too many endpoint')

        for version, uri in self.config['endpoint'].items():
            e = parse_grpc_endpoint(uri)
            self.client = pygrpc.client(endpoint=e['endpoint'], ssl_enabled=e['ssl_enabled'])

    def get_user(self, user_id, domain_id):
        return MessageToDict(self.client.User.get({'user_id': user_id, 'domain_id': domain_id},
                                                  metadata=self.transaction.get_connection_meta()),
                             preserving_proto_field_name=True)

    def list_users(self, query, domain_id):
        return MessageToDict(self.client.User.list({'query': query, 'domain_id': domain_id},
                                                   metadata=self.transaction.get_connection_meta()),
                             preserving_proto_field_name=True)

    def get_project(self, project_id, domain_id):
        return MessageToDict(self.client.Project.get({'project_id': project_id, 'domain_id': domain_id},
                                                     metadata=self.transaction.get_connection_meta()),
                             preserving_proto_field_name=True)

    def list_projects(self, query, domain_id):
        return MessageToDict(self.client.Project.list({'query': query, 'domain_id': domain_id},
                                                      metadata=self.transaction.get_connection_meta()),
                             preserving_proto_field_name=True)

    def list_domains(self, query):
        return MessageToDict(self.client.Domain.list({'query': query},
                                                   metadata=self.transaction.get_connection_meta()),
                             preserving_proto_field_name=True)



from spaceone.core.service import *
from spaceone.inventory.error import *
from spaceone.inventory.manager.identity_manager import IdentityManager
from spaceone.inventory.manager.cloud_service_stats_manager import CloudServiceStatsManager


_KEYWORD_FILTER = ['query_set_id', 'name']


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class CloudServiceStatsService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cloud_svc_stats_mgr: CloudServiceStatsManager = self.locator.get_manager('CloudServiceStatsManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query_set_id', 'domain_id'])
    @append_query_filter(['query_set_id', 'provider', 'cloud_service_group', 'cloud_service_type',
                          'region_code', 'account', 'project_id', 'domain_id'])
    @append_keyword_filter(_KEYWORD_FILTER)
    @set_query_page_limit(1000)
    def list(self, params):
        """ List Cloud Service Stats
        Args:
            params (dict): {
                    'query_set_id': 'str',
                    'provider': 'str',
                    'cloud_service_group': 'str',
                    'cloud_service_type': 'str',
                    'region_code': 'str',
                    'account': 'str',
                    'project_id': 'str',
                    'domain_id': 'str',
                    'query': 'dict (spaceone.api.core.v1.Query)',
                    'user_projects': 'list', // from meta
                }

        Returns:
            results (list)
            total_count (int)

        """
        query = params.get('query', {})
        query = self._change_project_group_filter(query, params['domain_id'])

        return self.cloud_svc_stats_mgr.list_cloud_service_stats(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'query.granularity', 'query.start', 'query.end', 'query.fields',
                     'query_set_id', 'domain_id'])
    @append_query_filter(['query_set_id', 'domain_id'])
    @append_keyword_filter(_KEYWORD_FILTER)
    @set_query_page_limit(1000)
    def analyze(self, params):
        """ Analyze Cloud Service Statistics Data
        Args:
            params (dict): {
                'query_set_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.TimeSeriesAnalyzeQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """
        domain_id = params['domain_id']
        query_set_id = params['query_set_id']
        query = params.get('query', {})

        query = self._change_project_group_filter(query, domain_id)

        return self.cloud_svc_stats_mgr.analyze_cloud_service_stats_by_granularity(query, domain_id, query_set_id)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['query_set_id', 'domain_id'])
    @append_keyword_filter(_KEYWORD_FILTER)
    def stat(self, params):
        """ Get Cloud Service Statistics Data
        Args:
            params (dict): {
                'query_set_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        query = self._change_project_group_filter(query, params['domain_id'])

        return self.cloud_svc_stats_mgr.stat_cloud_service_stats(query)

    def _change_project_group_filter(self, query, domain_id):
        change_filter = []

        project_group_query = {
            'filter': [],
            'only': ['project_group_id']
        }

        for condition in query.get('filter', []):
            key = condition.get('key', condition.get('k'))
            value = condition.get('value', condition.get('v'))
            operator = condition.get('operator', condition.get('o'))

            if not all([key, operator]):
                raise ERROR_DB_QUERY(reason='filter condition should have key, value and operator.')

            if key == 'project_group_id':
                project_group_query['filter'].append(condition)
            else:
                change_filter.append(condition)

        if len(project_group_query['filter']) > 0:
            identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
            response = identity_mgr.list_project_groups(project_group_query, domain_id)
            project_group_ids = []
            project_ids = []
            for project_group_info in response.get('results', []):
                project_group_ids.append(project_group_info['project_group_id'])

            for project_group_id in project_group_ids:
                response = identity_mgr.list_projects_in_project_group(project_group_id, domain_id, True,
                                                                       {'only': ['project_id']})
                for project_info in response.get('results', []):
                    if project_info['project_id'] not in project_ids:
                        project_ids.append(project_info['project_id'])

            change_filter.append({'k': 'project_id', 'v': project_ids, 'o': 'in'})

        query['filter'] = change_filter
        return query



from spaceone.api.inventory.v1 import region_pb2, region_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class Region(BaseAPI, region_pb2_grpc.RegionServicer):

    pb2 = region_pb2
    pb2_grpc = region_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('RegionService', metadata) as region_service:
            return self.locator.get_info('RegionInfo', region_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('RegionService', metadata) as region_service:
            return self.locator.get_info('RegionInfo', region_service.update(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('RegionService', metadata) as region_service:
            region_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('RegionService', metadata) as region_service:
            return self.locator.get_info('RegionInfo', region_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('RegionService', metadata) as region_service:
            region_vos, total_count = region_service.list(params)
            return self.locator.get_info('RegionsInfo', region_vos, total_count, minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('RegionService', metadata) as region_service:
            return self.locator.get_info('StatisticsInfo', region_service.stat(params))



from spaceone.core import utils
from spaceone.core.error import *


class ResourceManager(object):

    resource_keys: list = None
    query_method = None

    """
    This is used by collector
    """
    def query_resources(self, query, change_rules):
        secrets = []
        only = ['collection_info.secrets']
        change_values = {}
        change_key_map = {}
        for rule in change_rules:
            resource_key = rule['resource_key']
            change_key = rule['change_key']

            change_key_map[resource_key] = change_key
            change_values[change_key] = []
            only.append(resource_key)

        vos, total_count = getattr(self, self.query_method)(query)

        for vo in vos:
            data = vo.to_dict()
            for resource_key, change_key in change_key_map.items():
                value = utils.get_dict_value(data, resource_key)
                if value:
                    change_values[change_key].append(value)

            secrets = secrets + utils.get_dict_value(data, 'collection_info.secrets', [])

        for key, values in change_values.items():
            change_values[key] = list(set(values))

        return change_values, list(set(secrets))

    def find_resources(self, query):
        self._check_resource_finder_state()
        query['only'] = self.resource_keys

        resources = []
        vos, total_count = getattr(self, self.query_method)(query)

        for vo in vos:
            data = {}
            for key in self.resource_keys:
                data[key] = getattr(vo, key)

            resources.append(data)

        return resources, total_count

    def update_collection_state(self, query, state):
        self._check_resource_finder_state()
        query['only'] = self.resource_keys + ['collection_info', 'updated_at']

        resources = []
        vos, total_count = getattr(self, self.query_method)(query)

        for vo in vos:
            data = {}
            for key in self.resource_keys:
                data[key] = getattr(vo, key)
            try:
                self._update_collection_state_by_vo(vo, state)
            except Exception as e:
                pass
            resources.append(data)

        return resources, total_count

    def delete_resources(self, query):
        self._check_resource_finder_state()
        query['only'] = self.resource_keys + ['updated_at']

        vos, total_count = getattr(self, self.query_method)(query)
        vos.delete()

        return total_count

    def _update_collection_state_by_vo(self, resource_vo, state):
        def _rollback(old_data):
            print(f'[ROLLBACK] Revert Data : {old_data["name"]})')
            resource_vo.update(old_data)

        self.transaction.add_rollback(_rollback, resource_vo.to_dict())
        params = resource_vo.to_dict()['collection_info']
        params.update({'state': state})
        if state == "DISCONNECTED" or state == "DELETED":
            return resource_vo.update({'collection_info': params, 'updated_at': resource_vo.updated_at})
        else:
            return resource_vo.update({'collection_info': params})

    def _check_resource_finder_state(self):
        if not (self.resource_keys and self.query_method):
            raise ERROR_UNKNOWN(message='ResourceManager is not set.')

        if getattr(self, self.query_method, None) is None:
            raise ERROR_UNKNOWN(message='ResourceManager is not set.')



"""
Deprecated:
  Not used. Integrated with SpaceConnector.
"""

import logging

from google.protobuf.json_format import MessageToDict

from spaceone.core.connector import BaseConnector
from spaceone.core import pygrpc
from spaceone.core.utils import parse_grpc_endpoint
from spaceone.core.error import *


__all__ = ['ConfigConnector']
_LOGGER = logging.getLogger(__name__)


class ConfigConnector(BaseConnector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'endpoint' not in self.config:
            raise ERROR_WRONG_CONFIGURATION(key='endpoint')

        if len(self.config['endpoint']) > 1:
            raise ERROR_WRONG_CONFIGURATION(key='too many endpoint')

        for version, uri in self.config['endpoint'].items():
            e = parse_grpc_endpoint(uri)
            self.client = pygrpc.client(endpoint=e['endpoint'], ssl_enabled=e['ssl_enabled'])

    def get_domain_config(self, name, domain_id):
        return MessageToDict(
            self.client.DomainConfig.get(
                {'name': name, 'domain_id': domain_id},
                metadata=self.transaction.get_connection_meta()
            ),
            preserving_proto_field_name=True
        )



import logging
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector

__ALL__ = ['PluginManager']

_LOGGER = logging.getLogger(__name__)


class PluginManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.plugin_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='plugin')

    def get_endpoint(self, plugin_id, version, domain_id, upgrade_mode='AUTO'):
        response = self.plugin_connector.dispatch('Plugin.get_plugin_endpoint', {
            'plugin_id': plugin_id,
            'version': version,
            'upgrade_mode': upgrade_mode,
            'domain_id': domain_id
        })

        return response.get('endpoint'), response.get('updated_version')



import copy
import logging

from spaceone.core import utils
from spaceone.core.manager import BaseManager
from spaceone.inventory.model.cloud_service_type_model import CloudServiceType
from spaceone.inventory.lib.resource_manager import ResourceManager
from spaceone.inventory.manager.cloud_service_query_set_manager import CloudServiceQuerySetManager


_LOGGER = logging.getLogger(__name__)


class CloudServiceTypeManager(BaseManager, ResourceManager):

    resource_keys = ['cloud_service_type_id']
    query_method = 'list_cloud_service_types'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cloud_svc_type_model: CloudServiceType = self.locator.get_model('CloudServiceType')
        self.cloud_svc_query_set_mgr: CloudServiceQuerySetManager = self.locator.get_manager('CloudServiceQuerySetManager')

    def create_cloud_service_type(self, params):
        def _rollback(cloud_svc_type_vo):
            _LOGGER.info(f'[ROLLBACK] Delete Cloud Service Type : {cloud_svc_type_vo.name} ({cloud_svc_type_vo.cloud_service_type_id})')
            cloud_svc_type_vo.delete(True)

        cloud_svc_type_vo: CloudServiceType = self.cloud_svc_type_model.create(params)
        self.transaction.add_rollback(_rollback, cloud_svc_type_vo)

        self._create_cloud_service_query_sets(params.get('metadata', {}), cloud_svc_type_vo)

        return cloud_svc_type_vo

    def update_cloud_service_type(self, params):
        cloud_service_type_vo = self.get_cloud_service_type(params['cloud_service_type_id'], params['domain_id'])
        return self.update_cloud_service_type_by_vo(params, cloud_service_type_vo)

    def update_cloud_service_type_by_vo(self, params, cloud_svc_type_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[ROLLBACK] Revert Data : {old_data.get("cloud_service_type_id")}')
            cloud_svc_type_vo.update(old_data)

        self.transaction.add_rollback(_rollback, cloud_svc_type_vo.to_dict())

        self._update_cloud_service_query_sets(params.get('metadata', {}), cloud_svc_type_vo)

        return cloud_svc_type_vo.update(params)

    def delete_cloud_service_type(self, cloud_service_type_id, domain_id):
        cloud_svc_type_vo = self.get_cloud_service_type(cloud_service_type_id, domain_id)
        query_set_vos = self._filter_cloud_service_query_sets(cloud_svc_type_vo)
        query_set_vos.delete()

        self.delete_cloud_service_type_by_vo(cloud_svc_type_vo)

    def get_cloud_service_type(self, cloud_service_type_id, domain_id, only=None):
        return self.cloud_svc_type_model.get(cloud_service_type_id=cloud_service_type_id, domain_id=domain_id,
                                             only=only)

    def list_cloud_service_types(self, query):
        cloud_svc_type_vos, total_count = self.cloud_svc_type_model.query(**query)
        return cloud_svc_type_vos, total_count

    def stat_cloud_service_types(self, query):
        return self.cloud_svc_type_model.stat(**query)

    def delete_cloud_service_type_by_vo(self, cloud_svc_type_vo):
        self._delete_cloud_service_query_sets(cloud_svc_type_vo)
        cloud_svc_type_vo.delete()

    def delete_resources(self, query):
        query['only'] = self.resource_keys

        cloud_service_type_vos, total_count = self.list_cloud_service_types(query)
        for cloud_service_type_vo in cloud_service_type_vos:
            self.delete_cloud_service_type_by_vo(cloud_service_type_vo)

        return total_count

    def _create_cloud_service_query_sets(self, metadata: dict, cloud_service_type_vo: CloudServiceType):
        for query_set in metadata.get('query_sets', []):
            if 'name' in query_set:
                create_params = copy.deepcopy(query_set)
                self._create_cloud_service_query_set(create_params, cloud_service_type_vo)

    def _update_cloud_service_query_sets(self, metadata: dict, cloud_service_type_vo: CloudServiceType):
        query_set_vos = self._filter_cloud_service_query_sets(cloud_service_type_vo)
        query_set_info = {}

        for query_set_vo in query_set_vos:
            query_set_info[query_set_vo.name] = query_set_vo.query_set_id

        for query_set in metadata.get('query_sets', []):
            if name := query_set.get('name'):
                if name in query_set_info:
                    update_params = copy.deepcopy(query_set)
                    query_options = update_params.get('query_options', {})
                    if query_set_vos[0].query_hash != utils.dict_to_hash(query_options):
                        self.cloud_svc_query_set_mgr.update_cloud_service_query_set_by_vo(update_params,
                                                                                          query_set_vos[0])
                else:
                    create_params = copy.deepcopy(query_set)
                    self._create_cloud_service_query_set(create_params, cloud_service_type_vo)

                del query_set_info[name]

        for query_set_id in query_set_info.values():
            self.cloud_svc_query_set_mgr.delete_cloud_service_query_set(query_set_id, cloud_service_type_vo.domain_id)

    def _delete_cloud_service_query_sets(self, cloud_service_type_vo: CloudServiceType):
        query_set_vos = self._filter_cloud_service_query_sets(cloud_service_type_vo)
        for query_set_vo in query_set_vos:
            self.cloud_svc_query_set_mgr.delete_cloud_service_query_set_by_vo(query_set_vo)

    def _create_cloud_service_query_set(self, create_params: dict, cloud_service_type_vo: CloudServiceType):
        create_params['query_type'] = 'MANAGED'
        create_params['provider'] = cloud_service_type_vo.provider
        create_params['cloud_service_group'] = cloud_service_type_vo.group
        create_params['cloud_service_type'] = cloud_service_type_vo.name
        create_params['domain_id'] = cloud_service_type_vo.domain_id
        self.cloud_svc_query_set_mgr.create_cloud_service_query_set(create_params)

    def _filter_cloud_service_query_sets(self, cloud_service_type_vo: CloudServiceType):
        filter_params = {
            'provider': cloud_service_type_vo.provider,
            'cloud_service_group': cloud_service_type_vo.group,
            'cloud_service_type': cloud_service_type_vo.name,
            'domain_id': cloud_service_type_vo.domain_id
        }

        return self.cloud_svc_query_set_mgr.filter_cloud_service_query_sets(**filter_params)



from spaceone.inventory.service.region_service import RegionService
from spaceone.inventory.service.collector_service import CollectorService
from spaceone.inventory.service.job_service import JobService
from spaceone.inventory.service.job_task_service import JobTaskService
from spaceone.inventory.service.cloud_service_service import CloudServiceService
from spaceone.inventory.service.cloud_service_type_service import CloudServiceTypeService
from spaceone.inventory.service.cloud_service_query_set_service import CloudServiceQuerySetService
from spaceone.inventory.service.cloud_service_stats_service import CloudServiceStatsService
from spaceone.inventory.service.cleanup_service import CleanupService
from spaceone.inventory.service.resource_group_service import ResourceGroupService
from spaceone.inventory.service.note_service import NoteService
from spaceone.inventory.service.change_history_service import ChangeHistoryService
from spaceone.inventory.service.collector_rule_service import CollectorRuleService



import logging

from spaceone.core.manager import BaseManager
from spaceone.core import config
from spaceone.core.connector.space_connector import SpaceConnector

_LOGGER = logging.getLogger(__name__)


class SecretManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector',
                                                                           service='secret',
                                                                           token=config.get_global('TOKEN'))

    def get_secret_data(self, secret_id, domain_id):
        return self.secret_connector.dispatch('Secret.get_data', {'secret_id': secret_id, 'domain_id': domain_id})

    def get_secret(self, secret_id, domain_id):
        return self.secret_connector.dispatch('Secret.get', {'secret_id': secret_id, 'domain_id': domain_id})

    def list_secrets(self, query, domain_id):
        return self.secret_connector.dispatch('Secret.list', {'query': query, 'domain_id': domain_id})



import logging

from spaceone.core.manager import BaseManager
from spaceone.inventory.model.record_model import Record

_LOGGER = logging.getLogger(__name__)


class RecordManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.record_model: Record = self.locator.get_model('Record')

    def create_record(self, params):
        def _rollback(record_vo: Record):
            _LOGGER.info(
                f'[ROLLBACK] Delete Record : {record_vo.record_id} ({record_vo.cloud_service_id})')
            record_vo.delete()

        record_vo: Record = self.record_model.create(params)
        self.transaction.add_rollback(_rollback, record_vo)

        return record_vo

    def delete_record(self, record_id, domain_id):
        record_vo = self.get_record(record_id, domain_id)
        record_vo.delete()

    def get_record(self, record_id, domain_id, only=None):
        return self.record_model.get(record_id=record_id, domain_id=domain_id, only=only)

    def filter_records(self, **conditions):
        return self.record_model.filter(**conditions)

    def list_records(self, query={}):
        return self.record_model.query(**query)

    def stat_records(self, query):
        return self.record_model.stat(**query)



from spaceone.core.error import *


class ERROR_NO_COLLECTOR(ERROR_BASE):
    _message = '{collector_id} does not exist in {domain_id}'


class ERROR_COLLECTOR_STATE(ERROR_BASE):
    _message = 'collector state is {state}'


class ERROR_INIT_PLUGIN_FAILURE(ERROR_BASE):
    _message = 'Fail to init plugin, params={params}'


class ERROR_VERIFY_PLUGIN_FAILURE(ERROR_BASE):
    _message = 'Fail to verify plugin, params={params}'


class ERROR_NO_PLUGIN_PARAMETER(ERROR_BASE):
    _message = 'parameter: {param} is required'


class ERROR_TOKEN_AUTHENTICATION_FAILURE(ERROR_BASE):
    _message = 'A access token or refresh token is invalid.'


class ERROR_AUTHENTICATION_FAILURE_PLUGIN(ERROR_BASE):
    _message = 'External plugin authentication exception. (plugin_error_message={message})'


class ERROR_JOB_STATE_CHANGE(ERROR_BASE):
    _message = 'Job {job_id} state change: {status} -> {action}'


class ERROR_JOB_TASK_STATE_CHANGE(ERROR_BASE):
    _message = 'Job task {job_task_id} state change: {status} -> {action}'


class ERROR_COLLECT_FILTER(ERROR_BASE):
    _message = 'collect failed, plugin_info: {plugin_info}, filter: {param}'


class ERROR_COLLECTOR_SECRET(ERROR_BASE):
    _message = 'collect failed, plugin_info: {plugin_info}, secret_id: {param}'


class ERROR_JOB_UPDATE(ERROR_BASE):
    _message = 'job update failed, param={param}'


class ERROR_COLLECTOR_COLLECTING(ERROR_BASE):
    _message = 'collecting failed, plugin_info: {plugin_info}, filter: {filter}'


class ERROR_COLLECT_CANCELED(ERROR_BASE):
    _message = 'collecting canceled, job_id: {job_id}'


class ERROR_UNSUPPORTED_RESOURCE_TYPE(ERROR_BASE):
    _message = 'collector can not find resource_type: {resource_type}'


class ERROR_UNSUPPORTED_FILTER_KEY(ERROR_BASE):
    _message = 'request unsupported filter_key {filter_key} : {filter_value}'


class ERROR_COLLECT_INITIALIZE(ERROR_BASE):
    _message = 'failed on stage {stage}, params: {params}'


class ERROR_INVALID_PLUGIN_VERSION(ERROR_INVALID_ARGUMENT):
    _message = 'Plugin version is invalid. (plugin_id = {plugin_id}, version = {version})'


class ERROR_NOT_ALLOWED_PLUGIN_ID(ERROR_INVALID_ARGUMENT):
    _message = 'Changing plugin_id is not allowed. (old_plugin_id = {old_plugin_id}, new_plugin_id = {new_plugin_id})'


class ERROR_WRONG_PLUGIN_SETTINGS(ERROR_BASE):
    _message = 'The plugin settings is incorrect. (key = {key})'


class ERROR_INVALID_PLUGIN_OPTIONS(ERROR_INTERNAL_API):
    _message = 'The options received from the plugin is invalid. (reason = {reason})'


class ERROR_RESOURCE_KEYS_NOT_DEFINED(ERROR_BASE):
    _message = '{resource_type} manager does not define resource_keys field'


class ERROR_TOO_MANY_MATCH(ERROR_BASE):
    _message = 'The same resource exists. (\n' \
               'match_key = {match_key},\n' \
               'matched_resources = {resources},\n' \
               'more = {more})'


class ERROR_UNSUPPORTED_SCHEDULE(ERROR_BASE):
    _message = 'supported schedules: {supported}, requested: {requested}'


class ERROR_NOT_ALLOWED_SECRET_ID(ERROR_BASE):
    _message = 'Not allowed secret_id: {secret_id}'



from spaceone.api.inventory.v1 import cloud_service_type_pb2, cloud_service_type_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class CloudServiceType(BaseAPI, cloud_service_type_pb2_grpc.CloudServiceTypeServicer):

    pb2 = cloud_service_type_pb2
    pb2_grpc = cloud_service_type_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceTypeService', metadata) as cloud_svc_type_service:
            return self.locator.get_info('CloudServiceTypeInfo', cloud_svc_type_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceTypeService', metadata) as cloud_svc_type_service:
            return self.locator.get_info('CloudServiceTypeInfo', cloud_svc_type_service.update(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceTypeService', metadata) as cloud_svc_type_service:
            cloud_svc_type_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceTypeService', metadata) as cloud_svc_type_service:
            return self.locator.get_info('CloudServiceTypeInfo', cloud_svc_type_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceTypeService', metadata) as cloud_svc_type_service:
            cloud_svc_type_vos, total_count = cloud_svc_type_service.list(params)
            return self.locator.get_info('CloudServiceTypesInfo',
                                         cloud_svc_type_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceTypeService', metadata) as cloud_svc_type_service:
            return self.locator.get_info('StatisticsInfo', cloud_svc_type_service.stat(params))



import abc
from spaceone.inventory.error import *


PENDING = 'PENDING'
CANCELED = 'CANCELED'
INPROGRESS = 'IN_PROGRESS'
SUCCESS = 'SUCCESS'
FAILURE = 'FAILURE'


class JobTaskState(metaclass=abc.ABCMeta):
    def __init__(self):
        self.handle()

    @abc.abstractmethod
    def handle(self):
        pass


class PendingState(JobTaskState):
    def handle(self):
        pass

    def __str__(self):
        return PENDING


class CanceledState(JobTaskState):
    def handle(self):
        pass

    def __str__(self):
        return CANCELED


class InprogressState(JobTaskState):
    def handle(self):
        pass

    def __str__(self):
        return INPROGRESS


class SuccessState(JobTaskState):
    def handle(self):
        pass

    def __str__(self):
        return SUCCESS


class FailureState(JobTaskState):
    def handle(self):
        pass

    def __str__(self):
        return FAILURE


STATE_DIC = {
    'PENDING': PendingState(),
    'CANCELED': CanceledState(),
    'IN_PROGRESS': InprogressState(),
    'SUCCESS': SuccessState(),
    'FAILURE': FailureState()
}


class JobTaskStateMachine(object):
    def __init__(self, job_task_vo):
        self.job_task_id = job_task_vo.job_task_id
        self._state = STATE_DIC[job_task_vo.status]

    def inprogress(self):
        if isinstance(self._state, (PendingState, InprogressState, SuccessState, FailureState)):
            self._state = InprogressState()
        else:
            raise ERROR_JOB_TASK_STATE_CHANGE(action='INPROGRESS', job_task_id=self.job_task_id, status=str(self._state))
        return self.get_state()

    def success(self):
        if isinstance(self._state, (InprogressState)):
            self._state = SuccessState()
        else:
            raise ERROR_JOB_TASK_STATE_CHANGE(action='SUCCESS', job_task_id=self.job_task_id, status=str(self._state))
        return self.get_state()

    def failure(self):
        self._state = FailureState()
        return self.get_state()

    def canceled(self):
        self._state = CanceledState()
        return self.get_state()

    def get_state(self):
        return str(self._state)



import datetime
from spaceone.core import config
from spaceone.core.token import _validate_token
from spaceone.core.auth.jwt.jwt_util import JWTUtil


def init_count():
    current_time = datetime.datetime.utcnow()
    return {
        'previous': current_time,  # Last check_count time
        'index': 0,  # index
        'hour': current_time.hour,  # previous hour
        'started_at': 0,  # start time of push_token
        'ended_at': 0  # end time of execution in this tick
    }


def update_token():
    token = config.get_global('TOKEN')
    if token == "":
        token = _validate_token(config.get_global('TOKEN_INFO'))
    return token


def get_domain_id_from_token(token):
    decoded_token = JWTUtil.unverified_decode(token)
    return decoded_token['did']


def get_interval_value(schedule_vo):
    return schedule_vo.schedule.interval



import logging
import fnmatch

from spaceone.core.service import *
from spaceone.inventory.error import *
from spaceone.inventory.manager.collector_manager import CollectorManager
from spaceone.inventory.manager.collector_rule_manager import CollectorRuleManager
from spaceone.inventory.manager.identity_manager import IdentityManager
from spaceone.inventory.model.collector_rule_model import CollectorRule

_LOGGER = logging.getLogger(__name__)

_SUPPORTED_CONDITION_KEYS = ['provider', 'region_code', 'product', 'account', 'usage_type', 'resource_group',
                             'resource', 'tags.<key>', 'additional_info.<key>']
_SUPPORTED_CONDITION_OPERATORS = ['eq', 'contain', 'not', 'not_contain']


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class CollectorRuleService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.collector_rule_mgr: CollectorRuleManager = self.locator.get_manager('CollectorRuleManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_id', 'conditions_policy', 'actions', 'domain_id'])
    def create(self, params):
        """Create Collector rule

        Args:
            params (dict): {
                'collector_id': 'str',
                'name': 'str',
                'conditions': 'list',
                'conditions_policy': 'str',
                'actions': 'dict',
                'options': 'dict',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            collector_rule_vo (object)
        """

        domain_id = params['domain_id']
        collector_id = params['collector_id']
        conditions = params.get('conditions', [])
        conditions_policy = params['conditions_policy']
        actions = params['actions']
        rule_type = params.get('rule_type', 'CUSTOM')

        if conditions_policy == 'ALWAYS':
            params['conditions'] = []
        else:
            if len(conditions) == 0:
                raise ERROR_REQUIRED_PARAMETER(key='conditions')
            else:
                self._check_conditions(conditions)

        self._check_actions(actions, domain_id)

        collector_mgr: CollectorManager = self.locator.get_manager('CollectorManager')
        collector_vo = collector_mgr.get_collector(collector_id, domain_id)

        params['collector'] = collector_vo
        params['order'] = self._get_highest_order(collector_id, rule_type, domain_id) + 1

        return self.collector_rule_mgr.create_collector_rule(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_rule_id', 'domain_id'])
    def update(self, params):
        """ Update collector rule
        Args:
            params (dict): {
                'collector_rule_id': 'str',
                'name': 'str',
                'conditions': 'list',
                'conditions_policy': 'list',
                'actions': 'dict',
                'options': 'dict'
                'tags': 'dict'
                'domain_id': 'str'
            }

        Returns:
            collector_rule_vo (object)
        """
        collector_rule_id = params['collector_rule_id']
        domain_id = params['domain_id']
        conditions_policy = params.get('conditions_policy')
        conditions = params.get('conditions', [])

        collector_rule_vo = self.collector_rule_mgr.get_collector_rule(collector_rule_id, domain_id)

        if collector_rule_vo.rule_type == 'MANAGED':
            raise ERROR_NOT_ALLOWED_TO_UPDATE_RULE()

        if conditions_policy:
            if conditions_policy == 'ALWAYS':
                params['conditions'] = []
            else:
                if len(conditions) == 0:
                    raise ERROR_REQUIRED_PARAMETER(key='conditions')
                else:
                    self._check_conditions(conditions)

        if 'actions' in params:
            self._check_actions(params['actions'], domain_id)

        return self.collector_rule_mgr.update_collector_rule_by_vo(params, collector_rule_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_rule_id', 'order', 'domain_id'])
    def change_order(self, params):
        """ Change collector rule's order

        Args:
            params (dict): {
                'collector_rule_id': 'str',
                'order': 'int',
                'domain_id': 'str'
            }

        Returns:
            collector_rule_vo (object)
        """
        collector_rule_id = params['collector_rule_id']
        order = params['order']
        domain_id = params['domain_id']

        self._check_order(order)

        target_collector_rule_vo: CollectorRule = self.collector_rule_mgr.get_collector_rule(collector_rule_id,
                                                                                             domain_id)

        if target_collector_rule_vo.rule_type == 'MANAGED':
            raise ERROR_NOT_ALLOWED_TO_CHANGE_ORDER()

        if target_collector_rule_vo.order == order:
            return target_collector_rule_vo

        highest_order = self._get_highest_order(target_collector_rule_vo.collector_id,
                                                target_collector_rule_vo.rule_type,
                                                target_collector_rule_vo.domain_id)

        if order > highest_order:
            raise ERROR_INVALID_PARAMETER(key='order',
                                          reason=f'There is no collector rules greater than the {str(order)} order.')

        collector_rule_vos = self._get_all_collector_rules(target_collector_rule_vo.collector_id,
                                                           target_collector_rule_vo.rule_type,
                                                           target_collector_rule_vo.domain_id,
                                                           target_collector_rule_vo.collector_rule_id)

        collector_rule_vos.insert(order - 1, target_collector_rule_vo)

        i = 0
        for collector_rule_vo in collector_rule_vos:
            if target_collector_rule_vo != collector_rule_vo:
                self.collector_rule_mgr.update_collector_rule_by_vo({'order': i + 1}, collector_rule_vo)

            i += 1

        return self.collector_rule_mgr.update_collector_rule_by_vo({'order': order}, target_collector_rule_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_rule_id', 'domain_id'])
    def delete(self, params):
        """ Delete collector rule

        Args:
            params (dict): {
                'collector_rule_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """
        collector_rule_id = params['collector_rule_id']
        domain_id = params['domain_id']

        collector_rule_vo: CollectorRule = self.collector_rule_mgr.get_collector_rule(collector_rule_id, domain_id)
        rule_type = collector_rule_vo.rule_type

        if rule_type == 'MANAGED':
            raise ERROR_NOT_ALLOWED_TO_DELETE_RULE()

        collector_id = collector_rule_vo.collector_id
        self.collector_rule_mgr.delete_collector_rule_by_vo(collector_rule_vo)

        collector_rule_vos = self._get_all_collector_rules(collector_id, rule_type, domain_id)

        i = 0
        for collector_rule_vo in collector_rule_vos:
            self.collector_rule_mgr.update_collector_rule_by_vo({'order': i + 1}, collector_rule_vo)
            i += 1

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_rule_id', 'domain_id'])
    def get(self, params):
        """ Get collector rule

        Args:
            params (dict): {
                'collector_rule_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            collector_rule_vo (object)
        """
        collector_rule_id = params['collector_rule_id']
        domain_id = params['domain_id']

        return self.collector_rule_mgr.get_collector_rule(collector_rule_id, domain_id, params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['collector_rule_id', 'name', 'collector_id', 'domain_id'])
    @append_keyword_filter(['collector_rule_id', 'name'])
    def list(self, params):
        """ List collector rule

        Args:
            params (dict): {
                'collector_rule_id': 'str',
                'name': 'str',
                'collector_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            collector_rule_vos (object)
            total_count
        """
        query = params.get('query', {})
        return self.collector_rule_mgr.list_collector_rules(query)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(['collector_rule_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list) : 'list of statistics data'

        """
        query = params.get('query', {})
        return self.collector_rule_mgr.stat_collector_rules(query)

    def _check_actions(self, actions, domain_id):
        if 'change_project' in actions:
            project_id = actions['change_project']

            identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
            identity_mgr.get_project(project_id, domain_id)

        if 'match_project' in actions:
            if 'source' not in actions['match_project']:
                raise ERROR_REQUIRED_PARAMETER(key='actions.match_project.source')

        if 'match_service_account' in actions:
            if 'source' not in actions['match_service_account']:
                raise ERROR_REQUIRED_PARAMETER(key='actions.match_service_account.source')

    def _get_highest_order(self, collector_id, rule_type, domain_id):
        collector_rule_vos = self.collector_rule_mgr.filter_collector_rules(collector_id=collector_id,
                                                                            rule_type=rule_type,
                                                                            domain_id=domain_id)

        return collector_rule_vos.count()

    def _get_all_collector_rules(self, collector_id, rule_type, domain_id, exclude_collector_rule_id=None):
        query = {
            'filter': [
                {
                    'k': 'domain_id',
                    'v': domain_id,
                    'o': 'eq'
                },
                {
                    'k': 'collector_id',
                    'v': collector_id,
                    'o': 'eq'
                },
                {
                    'k': 'rule_type',
                    'v': rule_type,
                    'o': 'eq'
                },
            ],
            'sort': {
                'key': 'order'
            }
        }

        if exclude_collector_rule_id is not None:
            query['filter'].append({
                'k': 'collector_rule_id',
                'v': exclude_collector_rule_id,
                'o': 'not'
            })

        collector_rule_vos, total_count = self.collector_rule_mgr.list_collector_rules(query)
        return list(collector_rule_vos)

    @staticmethod
    def _check_conditions(conditions):
        for condition in conditions:
            key = condition.get('key')
            value = condition.get('value')
            operator = condition.get('operator')

            if not (key and value and operator):
                raise ERROR_INVALID_PARAMETER(key='conditions', reason='Condition should have key, value and operator.')

            if key not in _SUPPORTED_CONDITION_KEYS:
                if not (fnmatch.fnmatch(key, 'additional_info.*') or fnmatch.fnmatch(key, 'tags.*')):
                    raise ERROR_INVALID_PARAMETER(key='conditions.key',
                                                  reason=f'Unsupported key. '
                                                         f'({" | ".join(_SUPPORTED_CONDITION_KEYS)})')
            if operator not in _SUPPORTED_CONDITION_OPERATORS:
                raise ERROR_INVALID_PARAMETER(key='conditions.operator',
                                              reason=f'Unsupported operator. '
                                                     f'({" | ".join(_SUPPORTED_CONDITION_OPERATORS)})')

    @staticmethod
    def _check_order(order):
        if order <= 0:
            raise ERROR_INVALID_PARAMETER(key='order', reason='The order must be greater than 0.')



from spaceone.core.error import *


class ERROR_NOT_ALLOW_PINNING_KEYS(ERROR_INVALID_ARGUMENT):
    _message = 'Pinning keys not allowed. (key={key})'


class ERROR_METADATA_DICT_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'The value of metadata\'s {key} must be a dict type.'


class ERROR_METADATA_LIST_VALUE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'The value of metadata\'s {key} must be a list type.'



from mongoengine import *
from datetime import datetime

from spaceone.core.model.mongo_model import MongoModel
from spaceone.inventory.model.reference_resource_model import ReferenceResource
from spaceone.inventory.model.cloud_service_type_model import CloudServiceType
from spaceone.inventory.model.region_model import Region
from spaceone.inventory.error import *


class CollectionInfo(EmbeddedDocument):
    provider = StringField(max_length=40)
    service_account_id = StringField(max_length=40)
    secret_id = StringField(max_length=40)
    collector_id = StringField(max_length=40)
    last_collected_at = DateTimeField(auto_now=True)

    def to_dict(self):
        return dict(self.to_mongo())


class CloudService(MongoModel):
    cloud_service_id = StringField(max_length=40, generate_id='cloud-svc', unique=True)
    name = StringField(default=None, null=True)
    state = StringField(max_length=20, choices=('ACTIVE', 'DISCONNECTED', 'DELETED'), default='ACTIVE')
    account = StringField(max_length=255, default=None, null=True)
    instance_type = StringField(max_length=255, default=None, null=True)
    instance_size = FloatField(max_length=255, default=None, null=True)
    ip_addresses = ListField(StringField(max_length=255), default=[])
    cloud_service_group = StringField(max_length=255)
    cloud_service_type = StringField(max_length=255)
    provider = StringField(max_length=255)
    ref_cloud_service_type = StringField(max_length=255)
    data = DictField()
    metadata = DictField()
    reference = EmbeddedDocumentField(ReferenceResource, default=ReferenceResource)
    tags = DictField()
    tag_keys = DictField()
    region_code = StringField(max_length=255, default=None, null=True)
    ref_region = StringField(max_length=255, default=None, null=True)
    project_id = StringField(max_length=255, default=None, null=True)
    domain_id = StringField(max_length=40)
    collection_info = ListField(EmbeddedDocumentField(CollectionInfo), default=[])
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)
    deleted_at = DateTimeField(default=None, null=True)

    meta = {
        'updatable_fields': [
            'name',
            'data',
            'state',
            'account',
            'instance_type',
            'instance_size',
            'ip_addresses',
            'metadata',
            'reference',
            'tags',
            'tag_keys',
            'project_id',
            'region_code',
            'cloud_service_group',
            'cloud_service_type',
            'collection_info',
            'updated_at',
            'deleted_at',
        ],
        'minimal_fields': [
            'cloud_service_id',
            'name',
            'cloud_service_group',
            'cloud_service_type',
            'provider',
            'reference.resource_id',
            'region_code',
            'project_id'
        ],
        'change_query_keys': {
            'user_projects': 'project_id',
            'ip_address': 'ip_addresses'
        },
        'reference_query_keys': {
            'ref_cloud_service_type': {
                'model': CloudServiceType,
                'foreign_key': 'ref_cloud_service_type'
            },
            'ref_region': {
                'model': Region,
                'foreign_key': 'ref_region'
            }
        },
        'indexes': [
            'reference.resource_id',
            'state',
            {
                "fields": ['domain_id', 'state', 'cloud_service_id'],
                "name": "COMPOUND_INDEX_FOR_GC_1"
            },
            {
                "fields": ['domain_id', 'state', 'updated_at'],
                "name": "COMPOUND_INDEX_FOR_GC_2"
            },
            {
                "fields": ['domain_id', 'state', '-deleted_at'],
                "name": "COMPOUND_INDEX_FOR_GC_3"
            },
            {
                "fields": ['domain_id', 'state', 'provider', 'cloud_service_group', 'cloud_service_type',
                           'reference.resource_id', 'account', 'cloud_service_id'],
                "name": "COMPOUND_INDEX_FOR_COLLECTOR"
            },
            {
                "fields": ['domain_id', 'state', 'provider', 'cloud_service_group', 'cloud_service_type',
                           'project_id', 'region_code'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_1"
            },
            {
                "fields": ['domain_id', 'state', 'ref_cloud_service_type', 'project_id', 'region_code'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_2"
            },
            {
                "fields": ['domain_id', 'state', '-created_at', 'project_id'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_3"
            },
            {
                "fields": ['domain_id', 'state', '-deleted_at', 'project_id'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_4"
            },
            {
                "fields": ['domain_id', 'state', 'provider', 'cloud_service_group', 'cloud_service_type',
                           'project_id', 'instance_type'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_5"
            },
            {
                "fields": ['domain_id', 'cloud_service_id', 'project_id'],
                "name": "COMPOUND_INDEX_FOR_GET"
            }
        ]
    }

    def update(self, data):
        if self.state == 'DELETED':
            raise ERROR_RESOURCE_ALREADY_DELETED(resource_type='CloudService', resource_id=self.cloud_service_id)

        return super().update(data)

    def delete(self):
        if self.state == 'DELETED':
            raise ERROR_RESOURCE_ALREADY_DELETED(resource_type='CloudService', resource_id=self.cloud_service_id)

        self.update({
            'state': 'DELETED',
            'deleted_at': datetime.utcnow()
        })



from mongoengine import *
from spaceone.core.model.mongo_model import MongoModel


class Region(MongoModel):
    region_id = StringField(max_length=40, generate_id='region', unique=True)
    name = StringField(max_length=255)
    region_key = StringField(max_length=255)
    region_code = StringField(max_length=255, unique_with=['provider', 'domain_id'])
    provider = StringField(max_length=255)
    ref_region = StringField(max_length=255)
    tags = DictField()
    domain_id = StringField(max_length=255)
    updated_by = StringField(default=None, null=True)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        'updatable_fields': [
            'name',
            'region_key',
            'tags',
            'updated_by',
            'updated_at'
        ],
        'minimal_fields': [
            'region_id',
            'name',
            'region_code',
            'provider'
        ],
        'ordering': [
            'name'
        ],
        'indexes': [
            {
                "fields": ['domain_id', 'region_id'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_1"
            },
            {
                "fields": ['domain_id', 'region_code', 'provider'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_2"
            },
            {
                "fields": ['domain_id', 'region_key'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_3"
            },
            {
                "fields": ['region_id', 'ref_region'],
                "name": "COMPOUND_INDEX_FOR_REF_1"
            },
            {
                "fields": ['region_code', 'provider', 'ref_region'],
                "name": "COMPOUND_INDEX_FOR_REF_2"
            },
        ]
    }



from spaceone.api.inventory.v1 import resource_group_pb2, resource_group_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class ResourceGroup(BaseAPI, resource_group_pb2_grpc.ResourceGroupServicer):

    pb2 = resource_group_pb2
    pb2_grpc = resource_group_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ResourceGroupService', metadata) as resource_group_service:
            return self.locator.get_info('ResourceGroupInfo', resource_group_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ResourceGroupService', metadata) as resource_group_service:
            return self.locator.get_info('ResourceGroupInfo', resource_group_service.update(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ResourceGroupService', metadata) as resource_group_service:
            resource_group_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ResourceGroupService', metadata) as resource_group_service:
            return self.locator.get_info('ResourceGroupInfo', resource_group_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ResourceGroupService', metadata) as resource_group_service:
            resource_group_vos, total_count = resource_group_service.list(params)
            return self.locator.get_info('ResourceGroupsInfo', resource_group_vos, total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ResourceGroupService', metadata) as resource_group_service:
            return self.locator.get_info('StatisticsInfo', resource_group_service.stat(params))



from spaceone.core.service import *
from spaceone.inventory.manager.record_manager import RecordManager
from spaceone.inventory.manager.cloud_service_manager import CloudServiceManager


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ChangeHistoryService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.record_mgr: RecordManager = self.locator.get_manager('RecordManager')
        self.cloud_svc_mgr: CloudServiceManager = self.locator.get_manager('CloudServiceManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['cloud_service_id', 'domain_id'])
    @change_only_key({'collector_info': 'collector'}, key_path='query.only')
    @append_query_filter(['record_id', 'cloud_service_id', 'action', 'user_id', 'collector_id', 'job_id',
                          'updated_by', 'domain_id'])
    @append_keyword_filter(['diff.key', 'diff.before', 'diff.after'])
    def list(self, params):
        """
        Args:
            params (dict): {
                    'record_id': 'str',
                    'cloud_service_id': 'str',
                    'action': 'str',
                    'user_id': 'dict',
                    'collector_id': 'str',
                    'job_id': 'str',
                    'updated_by': 'str',
                    'domain_id  ': 'str',
                    'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                    'user_projects': 'list', // from meta
                }

        Returns:
            results (list)
            total_count (int)

        """

        self._check_cloud_service(params)

        query = params.get('query', {})
        return self.record_mgr.list_records(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_projects'])
    @append_keyword_filter(['diff.key', 'diff.before', 'diff.after'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        self._check_cloud_service(params)

        query = params.get('query', {})
        return self.record_mgr.stat_records(query)

    def _check_cloud_service(self, params):
        cloud_service_id = params['cloud_service_id']
        domain_id = params['domain_id']
        user_projects = params.get('user_projects')
        self.cloud_svc_mgr.get_cloud_service(cloud_service_id, domain_id, user_projects=user_projects)



from spaceone.api.inventory.v1 import change_history_pb2, change_history_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class ChangeHistory(BaseAPI, change_history_pb2_grpc.ChangeHistoryServicer):

    pb2 = change_history_pb2
    pb2_grpc = change_history_pb2_grpc

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ChangeHistoryService', metadata) as ch_service:
            record_vos, total_count = ch_service.list(params)
            return self.locator.get_info('ChangeHistoryInfo', record_vos, total_count, minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ChangeHistoryService', metadata) as ch_service:
            return self.locator.get_info('StatisticsInfo', ch_service.stat(params))



from mongoengine import *


class ReferenceResource(EmbeddedDocument):
    resource_id = StringField(default=None, null=True)
    external_link = StringField(default=None, null=True)

    def to_dict(self):
        return self.to_mongo()



import logging

from spaceone.core.manager import BaseManager
from spaceone.inventory.model.note_model import Note

_LOGGER = logging.getLogger(__name__)


class NoteManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.note_model: Note = self.locator.get_model('Note')

    def create_note(self, params):
        def _rollback(note_vo):
            _LOGGER.info(f'[create_note._rollback] '
                         f'Delete note : {note_vo.note_id}')
            note_vo.delete()

        note_vo: Note = self.note_model.create(params)
        self.transaction.add_rollback(_rollback, note_vo)

        return note_vo

    def update_note(self, params):
        note_vo: Note = self.get_note(params['note_id'], params['domain_id'])
        return self.update_note_by_vo(params, note_vo)

    def update_note_by_vo(self, params, note_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_note_by_vo._rollback] Revert Data : '
                         f'{old_data["note_id"]}')
            note_vo.update(old_data)

        self.transaction.add_rollback(_rollback, note_vo.to_dict())

        return note_vo.update(params)

    def delete_note(self, note_id, domain_id):
        note_vo: Note = self.get_note(note_id, domain_id)
        note_vo.delete()

    def get_note(self, note_id, domain_id, only=None):
        return self.note_model.get(note_id=note_id, domain_id=domain_id, only=only)

    def filter_notes(self, **conditions):
        return self.note_model.filter(**conditions)

    def list_notes(self, query={}):
        return self.note_model.query(**query)

    def stat_notes(self, query):
        return self.note_model.stat(**query)



def add_domain_id(query, domain_id: str):
    q = {'k': 'domain_id', 'v': domain_id, 'o': 'eq'}
    query.append(q)
    return query


def add_k_v_eq(query, k, v):
    q = {'k': k, 'v': v, 'o': 'eq'}
    query.append(q)
    return query


def find_data(dic: dict, key: str) -> str:
    """
    find hierarchy data
    :param dic:
    e.g.
    {
      'data': {
        'vm': {
          'vm_id': 'i-1234'
        }
      }
    }
    :param key: e.g 'data.vm.vm_id'
    :return: a found value(e.g 'i-1234'), otherwise, None.
    """
    if not isinstance(dic, dict) or not isinstance(key, str):
        return None

    key_parsed = key.split('.', 1)
    if len(key_parsed) > 1:
        return find_data(dic.get(key_parsed[0], None), key_parsed[1])
    else:
        return dic.get(key_parsed[0], None)


def dict_key_int_parser(dic: dict):
    # change key to int type, if all keys are able to cast to int type. Otherwise, leave it as original type.
    try:
        dic = {int(k): v for k, v in dic.items()}
    except Exception as e:
        pass
    return dic


def make_query(key, rules, resource, domain_id):
    query = []
    query = add_domain_id(query, domain_id)
    for rule in rules[key]:
        v = find_data(resource, rule)
        if v:
            query = add_k_v_eq(query, rule, v)

    query = {'filter': query}
    return query



import logging
import functools
from spaceone.core import utils
from spaceone.core.manager import BaseManager
from spaceone.inventory.manager.identity_manager import IdentityManager
from spaceone.inventory.model.collector_rule_model import CollectorRule, CollectorRuleCondition

_LOGGER = logging.getLogger(__name__)


class CollectorRuleManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.collector_rule_model: CollectorRule = self.locator.get_model('CollectorRule')
        self.identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
        self._project_info = {}
        self._service_account_info = {}
        self._collector_rule_info = {}

    def create_collector_rule(self, params):
        def _rollback(collector_rule_vo: CollectorRule):
            _LOGGER.info(f'[create_collector_rule._rollback] Delete event rule : {collector_rule_vo.name} '
                         f'({collector_rule_vo.collector_rule_id})')
            collector_rule_vo.delete()

        collector_rule_vo: CollectorRule = self.collector_rule_model.create(params)
        self.transaction.add_rollback(_rollback, collector_rule_vo)

        return collector_rule_vo

    def update_collector_rule(self, params):
        collector_rule_vo: CollectorRule = self.get_collector_rule(params['collector_rule_id'], params['domain_id'])
        return self.update_collector_rule_by_vo(params, collector_rule_vo)

    def update_collector_rule_by_vo(self, params, collector_rule_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_collector_rule_by_vo._rollback] Revert Data : '
                         f'{old_data["collector_rule_id"]}')
            collector_rule_vo.update(old_data)

        self.transaction.add_rollback(_rollback, collector_rule_vo.to_dict())

        return collector_rule_vo.update(params)

    def delete_collector_rule(self, collector_rule_id, domain_id):
        collector_rule_vo: CollectorRule = self.get_collector_rule(collector_rule_id, domain_id)
        self.delete_collector_rule_by_vo(collector_rule_vo)

    def get_collector_rule(self, collector_rule_id, domain_id, only=None):
        return self.collector_rule_model.get(collector_rule_id=collector_rule_id, domain_id=domain_id, only=only)

    def filter_collector_rules(self, **conditions):
        return self.collector_rule_model.filter(**conditions)

    def list_collector_rules(self, query={}):
        return self.collector_rule_model.query(**query)

    def stat_collector_rules(self, query):
        return self.collector_rule_model.stat(**query)

    def change_cloud_service_data(self, collector_id, domain_id, cloud_service_data):
        managed_collector_rule_vos, custom_collector_rule_vos = self._get_collector_rules(collector_id, domain_id)

        cloud_service_data = self._apply_collector_rule_to_cloud_service_data(cloud_service_data,
                                                                              managed_collector_rule_vos,
                                                                              domain_id)

        cloud_service_data = self._apply_collector_rule_to_cloud_service_data(cloud_service_data,
                                                                              custom_collector_rule_vos,
                                                                              domain_id)

        return cloud_service_data

    def _apply_collector_rule_to_cloud_service_data(self, cloud_service_data, collector_rule_vos, domain_id):
        for collector_rule_vo in collector_rule_vos:
            is_match = self._change_cloud_service_data_by_rule(cloud_service_data, collector_rule_vo)

            if is_match:
                cloud_service_data = self._change_cloud_service_data_with_actions(cloud_service_data,
                                                                                  collector_rule_vo.actions,
                                                                                  domain_id)

            if is_match and collector_rule_vo.options.stop_processing:
                break

        return cloud_service_data

    def _change_cloud_service_data_with_actions(self, cloud_service_data, actions, domain_id):
        for action, value in actions.items():
            if action == 'change_project':
                cloud_service_data['project_id'] = value

            elif action == 'match_project':
                source = value['source']
                target_key = value.get('target', 'project_id')
                target_value = utils.get_dict_value(cloud_service_data, source)

                if target_value:
                    project_info = self._get_project(target_key, target_value, domain_id)

                    if project_info:
                        cloud_service_data['project_id'] = project_info['project_id']

            elif action == 'match_service_account':
                source = value['source']
                target_key = value.get('target', 'service_account_id')
                target_value = utils.get_dict_value(cloud_service_data, source)
                if target_value:
                    service_account_info = self._get_service_account(target_key, target_value, domain_id)
                    if service_account_info:
                        cloud_service_data['service_account_id'] = service_account_info['service_account_id']
                        cloud_service_data['project_id'] = service_account_info.get('project_info', {}).get('project_id')

        return cloud_service_data

    def _get_service_account(self, target_key, target_value, domain_id):
        if f'service-account:{domain_id}:{target_key}:{target_value}' in self._service_account_info:
            return self._service_account_info[f'service-account:{domain_id}:{target_key}:{target_value}']

        query = {
            'filter': [
                {'k': target_key, 'v': target_value, 'o': 'eq'},
                {'k': 'service_account_type', 'v': 'GENERAL', 'o': 'eq'}
            ],
            'only': ['service_account_id', 'project_info']
        }

        response = self.identity_mgr.list_service_accounts(query, domain_id)
        results = response.get('results', [])
        total_count = response.get('total_count', 0)

        service_account_info = None
        if total_count > 0:
            service_account_info = results[0]

        self._service_account_info[f'service-account:{domain_id}:{target_key}:{target_value}'] = service_account_info
        return service_account_info

    def _get_project(self, target_key, target_value, domain_id):
        if f'project:{domain_id}:{target_key}:{target_value}' in self._project_info:
            _LOGGER.debug(f'[_get_project] Hit the cache in self._project_info : {target_value}')
            return self._project_info[f'project:{domain_id}:{target_key}:{target_value}']

        query = {
            'filter': [
                {'k': target_key, 'v': target_value, 'o': 'eq'}
            ],
            'only': ['project_id']
        }

        query_hash = utils.dict_to_hash(query)
        response = self.identity_mgr.list_projects_with_cache(query, query_hash, domain_id)
        results = response.get('results', [])
        total_count = response.get('total_count', 0)

        project_info = None
        if total_count > 0:
            project_info = results[0]

        self._project_info[f'project:{domain_id}:{target_key}:{target_value}'] = project_info
        return project_info

    def _change_cloud_service_data_by_rule(self, cloud_service_data, collector_rule_vo: CollectorRule):
        conditions_policy = collector_rule_vo.conditions_policy

        if conditions_policy == 'ALWAYS':
            return True
        else:
            results = list(map(functools.partial(self._check_condition, cloud_service_data), collector_rule_vo.conditions))

            if conditions_policy == 'ALL':
                return all(results)
            else:
                return any(results)

    @staticmethod
    def delete_collector_rule_by_vo(collector_rule_vo):
        collector_rule_vo.delete()

    @staticmethod
    def _check_condition(cloud_service_data, condition: CollectorRuleCondition):
        cloud_service_value = utils.get_dict_value(cloud_service_data, condition.key)
        condition_value = condition.value
        operator = condition.operator

        if cloud_service_value is None:
            return False

        if operator == 'eq':
            if cloud_service_value == condition_value:
                return True
            else:
                return False
        elif operator == 'contain':
            if cloud_service_value.lower().find(condition_value.lower()) >= 0:
                return True
            else:
                return False
        elif operator == 'not':
            if cloud_service_value != condition_value:
                return True
            else:
                return False
        elif operator == 'not_contain':
            if cloud_service_value.lower().find(condition_value.lower()) < 0:
                return True
            else:
                return False

        return False

    def _get_collector_rules(self, collector_id, domain_id):
        if collector_id in self._collector_rule_info:
            return self._collector_rule_info[collector_id].get('managed', []), \
                   self._collector_rule_info[collector_id].get('custom', [])

        managed_query = self._make_collector_rule_query(collector_id, 'MANAGED', domain_id)
        managed_collector_rule_vos, total_count = self.list_collector_rules(managed_query)

        custom_query = self._make_collector_rule_query(collector_id, 'CUSTOM', domain_id)
        custom_collector_rule_vos, total_count = self.list_collector_rules(custom_query)

        self._collector_rule_info[collector_id] = {}
        self._collector_rule_info[collector_id]['managed'] = managed_collector_rule_vos
        self._collector_rule_info[collector_id]['custom'] = custom_collector_rule_vos

        return managed_collector_rule_vos, custom_collector_rule_vos

    @staticmethod
    def _make_collector_rule_query(collector_id, rule_type, domain_id):
        return {
            'filter': [
                {'k': 'collector_id', 'v': collector_id, 'o': 'eq'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
                {'k': 'rule_type', 'v': rule_type, 'o': 'eq'}
            ],
            'sort': {'key': 'order'}
        }



######################################################################
#    ************ Very Important ************
#
# This is resource map for collector
# If you add new service and manager for specific RESOURCE_TYPE,
# add here for collector
######################################################################

RESOURCE_MAP = {
    'inventory.Server': ('ServerService', 'ServerManager'),                                 # Deprecated
    'inventory.FilterCache': ('CollectorService', 'FilterManager'),                         # Deprecated
    'inventory.CloudService': ('CloudServiceService', 'CloudServiceManager'),
    'inventory.CloudServiceType': ('CloudServiceTypeService', 'CloudServiceTypeManager'),
    'inventory.Region': ('RegionService', 'RegionManager'),
    'inventory.ErrorResource': ('CollectorService', 'CollectingManager'),
}


OP_MAP = {
    '=': 'eq',
    '>=': 'gte',
    '<=': 'lte',
    '>': 'gt',
    '<': 'lt',
    '!=': 'not'
}

DB_QUEUE_NAME = 'db_q'

NOT_COUNT = 0
CREATED = 1
UPDATED = 2
ERROR = 3

JOB_TASK_STAT_EXPIRE_TIME = 3600  # 1 hour
WATCHDOG_WAITING_TIME = 30  # wait 30 seconds, before watchdog works

MAX_MESSAGE_LENGTH = 2000



from spaceone.inventory.error.region import *
from spaceone.inventory.error.server import *
from spaceone.inventory.error.collector import *
from spaceone.inventory.error.collect_data import *
from spaceone.inventory.error.cloud_service_type import *
from spaceone.inventory.error.cloud_service_query_set import *
from spaceone.inventory.error.cloud_service_stats import *
from spaceone.inventory.error.collector_rule import *



import logging
from datetime import datetime
from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.inventory.error import *
from spaceone.inventory.model.collector_model import Collector


__ALL__ = ['CollectorManager']

_LOGGER = logging.getLogger(__name__)


class CollectorManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.collector_model: Collector = self.locator.get_model('Collector')

    def create_collector(self, params):
        """
        Args: params
          - name
          - plugin_info
          - schedule
          - state
          - tags
          - domain_id
        """
        def _rollback(collector_vo):
            _LOGGER.info(f'[ROLLBACK] Delete collector : {collector_vo.name} ({collector_vo.collector_id})')
            collector_vo.delete()

        collector_vo: Collector = self.collector_model.create(params)
        self.transaction.add_rollback(_rollback, collector_vo)
        return collector_vo

    def delete_collector(self, collector_id, domain_id):
        collector_vo = self.collector_model.get(collector_id=collector_id, domain_id=domain_id)
        collector_vo.delete()

    def get_collector(self, collector_id, domain_id, only=None):
        return self.collector_model.get(collector_id=collector_id, domain_id=domain_id, only=only)

    def enable_collector(self, collector_id, domain_id):
        collector_vo: Collector = self.collector_model.get(collector_id=collector_id, domain_id=domain_id)
        return collector_vo.update({'state': 'ENABLED'})

    def disable_collector(self, collector_id, domain_id, plugin_init=True):
        collector_vo: Collector = self.collector_model.get(collector_id=collector_id, domain_id=domain_id)
        return collector_vo.update({'state': 'DISABLED'})

    def list_collectors(self, query):
        return self.collector_model.query(**query)

    def stat_collectors(self, query):
        return self.collector_model.stat(**query)

    def update_last_collected_time(self, collector_vo):
        params = {'last_collected_at': datetime.utcnow()}
        self.update_collector_by_vo(collector_vo, params)

    @staticmethod
    def update_collector_by_vo(collector_vo, params):
        return collector_vo.update(params)

    @staticmethod
    def get_queue_name(name='collect_queue'):
        try:
            return config.get_global(name)
        except Exception as e:
            _LOGGER.warning(f'[_get_queue_name] name: {name} is not configured')
            return None




import logging
import copy
from datetime import datetime, date
from dateutil.relativedelta import relativedelta

from spaceone.core.manager import BaseManager
from spaceone.core import utils, cache
from spaceone.inventory.error.cloud_service_stats import *
from spaceone.inventory.model.cloud_service_stats_model import CloudServiceStats, MonthlyCloudServiceStats, \
    CloudServiceStatsQueryHistory

_LOGGER = logging.getLogger(__name__)


class CloudServiceStatsManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cloud_svc_stats_model: CloudServiceStats = self.locator.get_model('CloudServiceStats')
        self.monthly_stats_model: MonthlyCloudServiceStats = self.locator.get_model('MonthlyCloudServiceStats')

    def create_cloud_service_stats(self, params, execute_rollback=True):
        def _rollback(cloud_svc_stats_vo: CloudServiceStats):
            _LOGGER.info(f'[create_cloud_service_stats_data._rollback] '
                         f'Delete stats data : {cloud_svc_stats_vo.query_set_id} '
                         f'({cloud_svc_stats_vo.key})')
            cloud_svc_stats_vo.delete()

        cloud_svc_stats_vo: CloudServiceStats = self.cloud_svc_stats_model.create(params)

        if execute_rollback:
            self.transaction.add_rollback(_rollback, cloud_svc_stats_vo)

        return cloud_svc_stats_vo

    def create_monthly_cloud_service_stats(self, params, execute_rollback=True):
        def _rollback(monthly_stats_model: MonthlyCloudServiceStats):
            _LOGGER.info(f'[create_monthly_cloud_service_stats._rollback] '
                         f'Delete stats data : {monthly_stats_model.query_set_id} '
                         f'({monthly_stats_model.key})')
            monthly_stats_model.delete()

        monthly_stats_vo: MonthlyCloudServiceStats = self.monthly_stats_model.create(params)

        if execute_rollback:
            self.transaction.add_rollback(_rollback, monthly_stats_vo)

        return monthly_stats_vo

    def delete_cloud_service_stats(self, query_set_id, domain_id):
        cloud_svc_stats_vos = self.filter_cloud_service_stats(query_set_id=query_set_id, domain_id=domain_id)
        cloud_svc_stats_vos.delete()

    def filter_cloud_service_stats(self, **conditions):
        return self.cloud_svc_stats_model.filter(**conditions)

    def filter_monthly_cloud_service_stats(self, **conditions):
        return self.monthly_stats_model.filter(**conditions)

    def list_cloud_service_stats(self, query):
        query = self._append_status_filter(query)

        return self.cloud_svc_stats_model.query(**query)

    def list_monthly_cloud_service_stats(self, query):
        query = self._append_status_filter(query)

        return self.monthly_stats_model.query(**query)

    def stat_cloud_service_stats(self, query):
        query = self._append_status_filter(query)

        return self.cloud_svc_stats_model.stat(**query)

    def stat_monthly_cloud_service_stats(self, query):
        query = self._append_status_filter(query)

        return self.monthly_stats_model.stat(**query)

    def analyze_cloud_service_stats(self, query, target='SECONDARY_PREFERRED'):
        query['target'] = target
        query['date_field'] = 'created_date'
        query['date_field_format'] = '%Y-%m-%d'
        _LOGGER.debug(f'[analyze_cloud_service_stats] query: {query}')
        return self.cloud_svc_stats_model.analyze(**query)

    def analyze_monthly_cloud_service_stats(self, query, target='SECONDARY_PREFERRED'):
        query['target'] = target
        query['date_field'] = 'created_month'
        query['date_field_format'] = '%Y-%m'
        _LOGGER.debug(f'[analyze_monthly_cloud_service_stats] query: {query}')
        return self.monthly_stats_model.analyze(**query)

    def analyze_yearly_cloud_service_stats(self, query, target='SECONDARY_PREFERRED'):
        query['target'] = target
        query['date_field'] = 'created_year'
        query['date_field_format'] = '%Y'
        _LOGGER.debug(f'[analyze_yearly_cloud_service_stats] query: {query}')
        return self.monthly_stats_model.analyze(**query)

    @cache.cacheable(key='inventory:cloud-service-stats:daily:{domain_id}:{query_set_id}:{query_hash}', expire=3600 * 24)
    def analyze_cloud_service_stats_with_cache(self, query, query_hash, domain_id, query_set_id, target='SECONDARY_PREFERRED'):
        return self.analyze_cloud_service_stats(query, target)

    @cache.cacheable(key='inventory:cloud-service-stats:monthly:{domain_id}:{query_set_id}:{query_hash}', expire=3600 * 24)
    def analyze_monthly_cloud_service_stats_with_cache(self, query, query_hash, domain_id, query_set_id,
                                                       target='SECONDARY_PREFERRED'):
        return self.analyze_monthly_cloud_service_stats(query, target)

    @cache.cacheable(key='inventory:cloud-service-stats:yearly:{domain_id}:{query_set_id}:{query_hash}', expire=3600 * 24)
    def analyze_yearly_cloud_service_stats_with_cache(self, query, query_hash, domain_id, query_set_id,
                                                      target='SECONDARY_PREFERRED'):
        return self.analyze_yearly_cloud_service_stats(query, target)

    def analyze_cloud_service_stats_by_granularity(self, query, domain_id, query_set_id):
        self._check_date_range(query)
        granularity = query['granularity']

        # Save query history to speed up data loading
        query_hash = utils.dict_to_hash(query)
        self.create_cloud_service_stats_query_history(query, query_hash, domain_id, query_set_id)

        if granularity == 'DAILY':
            response = self.analyze_cloud_service_stats_with_cache(query, query_hash, domain_id, query_set_id)
        elif granularity == 'MONTHLY':
            response = self.analyze_monthly_cloud_service_stats_with_cache(query, query_hash, domain_id, query_set_id)
        else:
            response = self.analyze_yearly_cloud_service_stats_with_cache(query, query_hash, domain_id, query_set_id)

        return response

    @cache.cacheable(key='inventory:stats-query-history:{domain_id}:{query_set_id}:{query_hash}', expire=600)
    def create_cloud_service_stats_query_history(self, query, query_hash, domain_id, query_set_id):
        def _rollback(history_vo):
            _LOGGER.info(f'[create_cloud_service_stats_query_history._rollback] Delete query history: {query_hash}')
            history_vo.delete()

        history_model: CloudServiceStatsQueryHistory = self.locator.get_model('CloudServiceStatsQueryHistory')

        history_vos = history_model.filter(query_hash=query_hash, domain_id=domain_id)
        if history_vos.count() == 0:
            history_vo = history_model.create({
                'query_hash': query_hash,
                'query_options': copy.deepcopy(query),
                'query_set_id': query_set_id,
                'domain_id': domain_id
            })

            self.transaction.add_rollback(_rollback, history_vo)
        else:
            history_vos[0].update({})

    def _check_date_range(self, query):
        start_str = query.get('start')
        end_str = query.get('end')
        granularity = query.get('granularity')

        start = self._parse_start_time(start_str, granularity)
        end = self._parse_end_time(end_str, granularity)
        now = datetime.utcnow().date()

        if len(start_str) != len(end_str):
            raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                           reason='Start date and end date must be the same format.')

        if start >= end:
            raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                           reason='End date must be greater than start date.')

        if granularity == 'DAILY':
            if start + relativedelta(months=1) < end:
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='Request up to a maximum of 1 month.')

            if start + relativedelta(months=12) < now.replace(day=1):
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='For DAILY, you cannot request data older than 1 year.')

        elif granularity == 'MONTHLY':
            if start + relativedelta(months=12) < end:
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='Request up to a maximum of 12 months.')

            if start + relativedelta(months=36) < now.replace(day=1):
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='For MONTHLY, you cannot request data older than 3 years.')
        elif granularity == 'YEARLY':
            if start + relativedelta(years=3) < now.replace(month=1, day=1):
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='For YEARLY, you cannot request data older than 3 years.')

    def _parse_start_time(self, date_str, granularity):
        return self._convert_date_from_string(date_str.strip(), 'start', granularity)

    def _parse_end_time(self, date_str, granularity):
        end = self._convert_date_from_string(date_str.strip(), 'end', granularity)

        if granularity == 'YEARLY':
            return end + relativedelta(years=1)
        elif granularity == 'MONTHLY':
            return end + relativedelta(months=1)
        else:
            return end + relativedelta(days=1)

    @staticmethod
    def _convert_date_from_string(date_str, key, granularity):
        if granularity == 'YEARLY':
            date_format = '%Y'
            date_type = 'YYYY'
        elif granularity == 'MONTHLY':
            if len(date_str) == 4:
                date_format = '%Y'
                date_type = 'YYYY'
            else:
                date_format = '%Y-%m'
                date_type = 'YYYY-MM'
        else:
            if len(date_str) == 4:
                date_format = '%Y'
                date_type = 'YYYY'
            elif len(date_str) == 7:
                date_format = '%Y-%m'
                date_type = 'YYYY-MM'
            else:
                date_format = '%Y-%m-%d'
                date_type = 'YYYY-MM-DD'

        try:
            return datetime.strptime(date_str, date_format).date()
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=date_type)

    @staticmethod
    def _append_status_filter(query):
        query_filter = query.get('filter', [])
        query_filter.append({
            'k': 'status',
            'v': 'DONE',
            'o': 'eq'
        })
        query['filter'] = query_filter
        return query



import functools
from spaceone.api.inventory.v1 import cloud_service_type_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.inventory.model.cloud_service_type_model import CloudServiceType

__all__ = ['CloudServiceTypeInfo', 'CloudServiceTypesInfo']


def CloudServiceTypeInfo(cloud_svc_type_vo: CloudServiceType, minimal=False):
    info = {
        'cloud_service_type_id': cloud_svc_type_vo.cloud_service_type_id,
        'name': cloud_svc_type_vo.name,
        'provider': cloud_svc_type_vo.provider,
        'group': cloud_svc_type_vo.group,
        'service_code': cloud_svc_type_vo.service_code,
        'is_primary': cloud_svc_type_vo.is_primary,
        'is_major': cloud_svc_type_vo.is_major,
        'resource_type': cloud_svc_type_vo.resource_type
    }

    if not minimal:
        info.update({
            'cloud_service_type_key': cloud_svc_type_vo.cloud_service_type_key,
            'metadata': change_struct_type(cloud_svc_type_vo.metadata),
            'labels': change_list_value_type(cloud_svc_type_vo.labels),
            'tags': change_struct_type(cloud_svc_type_vo.tags),
            'domain_id': cloud_svc_type_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(cloud_svc_type_vo.created_at),
            'updated_at': utils.datetime_to_iso8601(cloud_svc_type_vo.updated_at)
        })

    return cloud_service_type_pb2.CloudServiceTypeInfo(**info)


def CloudServiceTypesInfo(cloud_svc_type_vos, total_count, **kwargs):
    return cloud_service_type_pb2.CloudServiceTypesInfo(results=list(map(functools.partial(CloudServiceTypeInfo, **kwargs),
                                                                         cloud_svc_type_vos)),
                                                        total_count=total_count)



import logging
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector

__ALL__ = ['CollectorPluginManager']

_LOGGER = logging.getLogger(__name__)


class CollectorPluginManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init_plugin(self, endpoint, options):
        plugin_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', endpoint=endpoint)
        return plugin_connector.dispatch('Collector.init', {'options': options})

    def verify_plugin(self, endpoint, options, secret_data):
        plugin_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', endpoint=endpoint)
        params = {'options': options, 'secret_data': secret_data}
        plugin_connector.dispatch('Collector.verify', params)

    def collect(self, endpoint, options, secret_data, collector_filter, task_options):
        plugin_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', endpoint=endpoint)

        params = {
            'options': options,
            'secret_data': secret_data,
            'filter': collector_filter
        }

        if task_options:
            params['task_options'] = task_options

        return plugin_connector.dispatch('Collector.collect', params)

    def get_tasks(self, endpoint, secret_data, options):
        plugin_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', endpoint=endpoint)

        params = {'options': options, 'secret_data': secret_data}
        return plugin_connector.dispatch('Job.get_tasks', params)



PROTO = {
    'spaceone.inventory.interface.grpc.v1.region': ['Region'],
    'spaceone.inventory.interface.grpc.v1.collector': ['Collector'],
    'spaceone.inventory.interface.grpc.v1.job': ['Job'],
    'spaceone.inventory.interface.grpc.v1.job_task': ['JobTask'],
    'spaceone.inventory.interface.grpc.v1.cloud_service_type': ['CloudServiceType'],
    'spaceone.inventory.interface.grpc.v1.cloud_service': ['CloudService'],
    'spaceone.inventory.interface.grpc.v1.resource_group': ['ResourceGroup'],
    'spaceone.inventory.interface.grpc.v1.change_history': ['ChangeHistory'],
    'spaceone.inventory.interface.grpc.v1.note': ['Note'],
    'spaceone.inventory.interface.grpc.v1.collector_rule': ['CollectorRule'],
    'spaceone.inventory.interface.grpc.v1.cloud_service_query_set': ['CloudServiceQuerySet'],
    'spaceone.inventory.interface.grpc.v1.cloud_service_stats': ['CloudServiceStats']
}



import functools
import logging
from spaceone.api.inventory.v1 import region_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.inventory.model.region_model import Region

__all__ = ['RegionInfo', 'RegionsInfo']

_LOGGER = logging.getLogger(__name__)


def RegionInfo(region_vo: Region, minimal=False):
    info = {
        'region_id': region_vo.region_id,
        'name': region_vo.name,
        'region_code': region_vo.region_code,
        'provider': region_vo.provider
    }

    if not minimal:
        info.update({
            'region_key': region_vo.region_key,
            'tags': change_struct_type(region_vo.tags),
            'domain_id': region_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(region_vo.created_at),
            'updated_at': utils.datetime_to_iso8601(region_vo.updated_at)
        })

    return region_pb2.RegionInfo(**info)


def RegionsInfo(region_vos, total_count, **kwargs):
    return region_pb2.RegionsInfo(results=list(map(functools.partial(RegionInfo, **kwargs), region_vos)),
                                  total_count=total_count)



import logging
from spaceone.core.service import *
from spaceone.core import utils
from spaceone.inventory.error import *
from spaceone.inventory.manager.region_manager import RegionManager

_LOGGER = logging.getLogger(__name__)
_KEYWORD_FILTER = ['region_id', 'name', 'region_code']


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class RegionService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.region_mgr: RegionManager = self.locator.get_manager('RegionManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    def create(self, params):
        """
        Args:
        params (dict): {
            'name': 'str',
            'region_code': 'str',
            'provider': 'str',
            'tags': 'list',
            'domain_id': 'str'
        }
        Returns:
            region_vo (object)
        """
        return self.create_resource(params)

    @check_required(['name', 'region_code', 'provider', 'domain_id'])
    def create_resource(self, params):

        if 'tags' in params:
            if isinstance(params['tags'], list):
                params['tags'] = utils.tags_to_dict(params['tags'])

        params['updated_by'] = self.transaction.get_meta('collector_id') or 'manual'
        params['region_key'] = f'{params["provider"]}.{params["region_code"]}'

        return self.region_mgr.create_region(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    def update(self, params):
        """
        Args:
        params (dict): {
            'region_id': 'str',
            'name': 'str',
            'tags': 'list',
            'domain_id': 'str'
        }
        Returns:
            region_vo (object)
        """
        return self.update_resource(params)

    @check_required(['region_id', 'domain_id'])
    def update_resource(self, params):
        if 'tags' in params:
            if isinstance(params['tags'], list):
                params['tags'] = utils.tags_to_dict(params['tags'])

        params['updated_by'] = self.transaction.get_meta('collector_id') or 'manual'

        region_vo = self.region_mgr.get_region(params['region_id'], params['domain_id'])

        if not region_vo.region_key:
            params['region_key'] = f'{region_vo.provider}.{region_vo.region_code}'

        return self.region_mgr.update_region_by_vo(params, region_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    def delete(self, params):
        """
        Args:
        params (dict): {
            'region_id': 'str',
            'domain_id': 'str'
        }
        Returns:
            None
        """
        self.delete_resource(params)

    @check_required(['region_id', 'domain_id'])
    def delete_resource(self, params):
        region_vo = self.region_mgr.get_region(params['region_id'], params['domain_id'])
        self.region_mgr.delete_region_by_vo(region_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['region_id', 'domain_id'])
    def get(self, params):
        """
        Args:
            params (dict): {
                    'region_id': 'str',
                    'domain_id': 'str',
                    'only': 'list'
                }

        Returns:
            region_vo (object)

        """
        return self.region_mgr.get_region(params['region_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['region_id', 'name', 'region_key', 'region_code', 'provider', 'domain_id'])
    @append_keyword_filter(_KEYWORD_FILTER)
    def list(self, params):
        """
        Args:
            params (dict): {
                    'region_id': 'str',
                    'name': 'str',
                    'region_key': 'str',
                    'region_code': 'str',
                    'provider': 'str',
                    'domain_id': 'str',
                    'query': 'dict (spaceone.api.core.v1.Query)'
                }

        Returns:
            results (list)
            total_count (int)

        """

        return self.region_mgr.list_regions(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(_KEYWORD_FILTER)
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.region_mgr.stat_regions(query)



DATABASE_AUTO_CREATE_INDEX = True
DATABASES = {
    'default': {
        'db': 'inventory',
        'host': 'localhost',
        'port': 27017,
        'username': '',
        'password': '',
        'read_preference': 'SECONDARY_PREFERRED'
    }
}

CACHES = {
    'default': {},
    'local': {
        'backend': 'spaceone.core.cache.local_cache.LocalCache',
        'max_size': 128,
        'ttl': 300
    }
}

# Garbage Collection Policies
JOB_TIMEOUT = 2  # 2 Hours
JOB_TERMINATION_TIME = 2 * 30  # 2 Months
RESOURCE_TERMINATION_TIME = 3 * 30  # 3 Months
DEFAULT_DELETE_POLICIES = {
    'inventory.CloudService': 48,  # 48 Hours
    'inventory.CloudServiceType': 3 * 30 * 24,  # 3 Months
    'inventory.Region': 48,  # 48 Hours
}
DEFAULT_DISCONNECTED_STATE_DELETE_POLICY = 3  # 3 Count
DELETE_EXCLUDE_DOMAINS = []

# Cloud Service Stats Schedule Settings
STATS_SCHEDULE_HOUR = 15  # Hour (UTC)

HANDLERS = {
}

CONNECTORS = {
    'CollectorPluginConnector': {
    },
    'SpaceConnector': {
        'backend': 'spaceone.core.connector.space_connector.SpaceConnector',
        'endpoints': {
            'identity': 'grpc://identity:50051',
            'plugin': 'grpc://plugin:50051',
            'repository': 'grpc://repository:50051',
            'secret': 'grpc://secret:50051',
            'config': 'grpc://config:50051',
        }
    },
}

ENDPOINTS = {}
LOG = {}
QUEUES = {}
SCHEDULERS = {}
WORKERS = {}
TOKEN = ""
TOKEN_INFO = {}
collect_queue = ""  # Queue name for asynchronous collect



import logging
from spaceone.core import cache, config
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector

_LOGGER = logging.getLogger(__name__)


class IdentityManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if token := kwargs.get('token'):
            self.identity_conn: SpaceConnector = self.locator.get_connector('SpaceConnector', service='identity',
                                                                            token=token)
        else:
            system_token = config.get_global('TOKEN')
            self.identity_conn: SpaceConnector = self.locator.get_connector('SpaceConnector', service='identity',
                                                                            token=system_token)

    def get_user(self, user_id, domain_id):
        return self.identity_conn.dispatch('User.get', {'user_id': user_id, 'domain_id': domain_id})

    def list_users(self, query, domain_id):
        return self.identity_conn.dispatch('User.list', {'query': query, 'domain_id': domain_id})

    @cache.cacheable(key='inventory:project:{domain_id}:{project_id}', expire=3600)
    def get_project(self, project_id, domain_id):
        _LOGGER.debug(f'[get_project] NO CACHE. Get projects - {project_id}')
        return self.identity_conn.dispatch('Project.get', {'project_id': project_id, 'domain_id': domain_id})

    def list_projects(self, query, domain_id):
        return self.identity_conn.dispatch('Project.list', {'query': query, 'domain_id': domain_id})

    @cache.cacheable(key='inventory:project:query:{domain_id}:{query_hash}', expire=3600)
    def list_projects_with_cache(self, query, query_hash, domain_id):
        _LOGGER.debug(f'[list_projects_with_cache] NO CACHE. Search list projects')
        return self.list_projects(query, domain_id)

    def list_project_groups(self, query, domain_id):
        return self.identity_conn.dispatch('ProjectGroup.list', {'query': query, 'domain_id': domain_id})

    def get_project_group(self, project_group_id, domain_id):
        return self.identity_conn.dispatch('ProjectGroup.get', {'project_group_id': project_group_id,
                                                                'domain_id': domain_id})

    def list_projects_in_project_group(self, project_group_id, domain_id, recursive=False, query=None):
        request = {
            'project_group_id': project_group_id,
            'domain_id': domain_id,
            'recursive': recursive
        }

        if query:
            request['query'] = query

        return self.identity_conn.dispatch('ProjectGroup.list_projects', request)

    def list_domains(self, query):
        return self.identity_conn.dispatch('Domain.list', {'query': query})

    def list_service_accounts(self, query, domain_id):
        return self.identity_conn.dispatch('ServiceAccount.list', {'query': query, 'domain_id': domain_id})



import logging
from operator import itemgetter
from spaceone.core.manager import BaseManager
from spaceone.core import utils
from spaceone.inventory.model.cloud_service_model import CloudService
from spaceone.inventory.manager.record_manager import RecordManager

_LOGGER = logging.getLogger(__name__)

DIFF_KEYS = [
    'name',
    'ip_addresses',
    'account',
    'instance_type',
    'instance_size',
    'reference',
    'region_code',
    'project_id',
    'data',
    'tags',
]

MAX_KEY_DEPTH = 3


class ChangeHistoryManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.record_mgr: RecordManager = self.locator.get_manager('RecordManager')
        self.merged_data = {}
        self.is_changed = False
        self.collector_id = self.transaction.get_meta('collector_id')
        self.job_id = self.transaction.get_meta('job_id')
        self.plugin_id = self.transaction.get_meta('plugin_id')
        self.secret_id = self.transaction.get_meta('secret.secret_id')
        self.service_account_id = self.transaction.get_meta('secret.service_account_id')
        self.user_id = self.transaction.get_meta('user_id')

        if self.collector_id and self.job_id and self.service_account_id and self.plugin_id:
            self.updated_by = 'COLLECTOR'
        else:
            self.updated_by = 'USER'

    def add_new_history(self, cloud_service_vo: CloudService, new_data: dict):
        self._create_record(cloud_service_vo, new_data)

    def add_update_history(self, cloud_service_vo: CloudService, new_data: dict, old_data: dict):
        new_keys = new_data.keys()

        if len(set(new_keys) & set(DIFF_KEYS)) > 0:
            self._create_record(cloud_service_vo, new_data, old_data)

    def add_delete_history(self, cloud_service_vo: CloudService):
        params = {
            'cloud_service_id': cloud_service_vo.cloud_service_id,
            'domain_id': cloud_service_vo.domain_id,
            'action': 'DELETE',
        }

        self.record_mgr.create_record(params)

    def _create_record(self, cloud_service_vo: CloudService, new_data, old_data=None):
        if old_data:
            action = 'UPDATE'
        else:
            action = 'CREATE'

        metadata = new_data.get('metadata', {}).get(self.plugin_id or 'MANUAL', {})
        exclude_keys = metadata.get('change_history', {}).get('exclude', [])

        diff = self._make_diff(new_data, old_data, exclude_keys)
        diff_count = len(diff)

        if diff_count > 0:
            params = {
                'cloud_service_id': cloud_service_vo.cloud_service_id,
                'domain_id': cloud_service_vo.domain_id,
                'action': action,
                'diff': diff,
                'diff_count': diff_count,
                'updated_by': self.updated_by,
            }

            if self.updated_by == 'COLLECTOR':
                params['collector_id'] = self.collector_id
                params['job_id'] = self.job_id
            else:
                params['user_id'] = self.user_id

            self.record_mgr.create_record(params)

    def _make_diff(self, new_data, old_data, exclude_keys):
        diff = []
        for key in DIFF_KEYS:
            if key in new_data:
                if old_data:
                    old_value = old_data.get(key)
                else:
                    old_value = None

                diff += self._get_diff_data(key, new_data[key], old_value, exclude_keys)

        return diff

    def _get_diff_data(self, key, new_value, old_value, exclude_keys, depth=1, parent_key=None):
        diff = []

        if depth == MAX_KEY_DEPTH:
            if new_value != old_value:
                diff_data = self._generate_diff_data(key, parent_key, new_value, old_value, exclude_keys)
                if diff_data:
                    diff.append(diff_data)
        elif isinstance(new_value, dict):
            if parent_key:
                parent_key = f'{parent_key}.{key}'
            else:
                parent_key = key

            for sub_key, sub_value in new_value.items():
                if isinstance(old_value, dict):
                    sub_old_value = old_value.get(sub_key)
                else:
                    sub_old_value = None

                diff += self._get_diff_data(sub_key, sub_value, sub_old_value, exclude_keys, depth+1, parent_key)
        else:
            if new_value != old_value:
                diff_data = self._generate_diff_data(key, parent_key, new_value, old_value, exclude_keys)
                if diff_data:
                    diff.append(diff_data)

        return diff

    def _generate_diff_data(self, key, parent_key, new_value, old_value, exclude_keys):
        if old_value is None:
            diff_type = 'ADDED'
        else:
            diff_type = 'CHANGED'

        before = self._change_diff_value(old_value)
        after = self._change_diff_value(new_value)
        diff_key = key if parent_key is None else f'{parent_key}.{key}'

        if diff_key in exclude_keys:
            return None
        elif before == after:
            return None
        else:
            return {
                'key': diff_key,
                'before': before,
                'after': after,
                'type': diff_type
            }

    def _change_diff_value(self, value):
        if isinstance(value, dict):
            return utils.dump_json(self._sort_dict_value(value))
        elif isinstance(value, list):
            return utils.dump_json(self._sort_list_values(value))
        elif value is None:
            return value
        else:
            return str(value)

    def _sort_dict_value(self, value: dict) -> dict:
        try:
            for k, v in value.items():
                if isinstance(v, dict):
                    value[k] = self._sort_dict_value(v)
                elif isinstance(v, list):
                    value[k] = self._sort_list_values(v)

            return dict(sorted(value.items()))
        except Exception as e:
            pass
            # _LOGGER.warning(f'[_sort_dict_value] dict value sort error: {e}', exc_info=True)

        return value

    def _sort_list_values(self, values: list) -> list:
        if len(values) > 0:
            if isinstance(values[0], dict):
                changed_list_values = []
                for value in values:
                    changed_list_values.append(self._sort_dict_value(value))

                sort_keys = list(changed_list_values[0].keys())

                if len(sort_keys) > 0:
                    try:
                        return sorted(changed_list_values, key=itemgetter(*sort_keys[:3]))
                    except Exception as e:
                        pass
                        # _LOGGER.warning(f'[_sort_list_values] list value sort error (sort_keys={sort_keys[:3]}, '
                        #                 f'reason={e})', exc_info=True)

                return changed_list_values

            else:
                return sorted(values)

        return values



import logging

from google.protobuf.json_format import MessageToDict

from spaceone.core import pygrpc
from spaceone.core.connector import BaseConnector
from spaceone.core.utils import parse_grpc_endpoint
from spaceone.inventory.error import *

_LOGGER = logging.getLogger(__name__)


class CollectorPluginConnector(BaseConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = None

    def initialize(self, endpoint):
        """ Initialize based on endpoint
        Args:
            endpoint (message): endpoint message
        """
        e = parse_grpc_endpoint(endpoint)
        self.client = pygrpc.client(endpoint=e['endpoint'], ssl_enabled=e['ssl_enabled'], max_message_length=1024*1024*32)

        if self.client is None:
            _LOGGER.error(f'[initialize] Cannot access gRPC server. '
                          f'(host: {e.get("hostname")}, port: {e.get("port")}, version: plugin)')
            raise ERROR_GRPC_CONFIGURATION(endpoint=endpoint)

    def init(self, options):
        params = {
            'options': options
        }
        try:
            meta = []
            plugin_info = self.client.Collector.init(params, metadata=meta)
            return MessageToDict(plugin_info)
        except Exception as e:
            _LOGGER.error(f'[init] error: {e}')
            raise ERROR_INIT_PLUGIN_FAILURE(params=params)

    def verify(self, options, secret_data):
        params = {
            'options': options,
            'secret_data': secret_data
            }
        try:
            # TODO: meta (plugin has no meta)
            meta = []
            verify_info = self.client.Collector.verify(params, metadata=meta)
            return MessageToDict(verify_info)
        except Exception as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(messsage=str(e))

    def collect(self, options, secret_data, filter, region_id=None):
        """ Collector Data base on param

        Unary/Stream Between This and Plugin

        return Job Result

        format of credential (for ec2 collector)
        secret_data = {
            'aws_access_key_id': 'MY_AWS_ACCESS_KEYID',
            'aws_secret_access_key': 'MY_AWS_SECRET_ACCESS_KEY',
            'region': 'ap-northeast-2'
        }
        """
        params = {
            'options': options,
            'secret_data': secret_data,
            'filter': filter
            }
        #params = {'options':{'domain':'mz.co.kr'},'credentials':{}}
        #_LOGGER.debug('[collect] correct params: %s' % params)
        # TODO: meta (plugin has no meta)
        meta = []
        result_stream = self.client.Collector.collect(params, metadata=meta)
        return result_stream



"""
Deprecated:
  Not used. Integrated with SpaceConnector.
"""

import logging

from spaceone.core import pygrpc
from spaceone.core.connector import BaseConnector
from spaceone.core.error import ERROR_WRONG_CONFIGURATION
from spaceone.core.utils import parse_grpc_endpoint

_LOGGER = logging.getLogger(__name__)


class PluginConnector(BaseConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._check_config()
        self._init_client()

    def _init_client(self):
        for version, uri in self.config['endpoint'].items():
            e = parse_grpc_endpoint(uri)
            self.client = pygrpc.client(endpoint=e['endpoint'], ssl_enabled=e['ssl_enabled'])

    def _check_config(self):
        _LOGGER.debug(f'[_check_config] config: {self.config}')
        if 'endpoint' not in self.config:
            raise ERROR_WRONG_CONFIGURATION(key='endpoint')
        if len(self.config['endpoint']) > 1:
            raise ERROR_WRONG_CONFIGURATION(key='too many endpoint')

    def get_plugin_endpoint(self, plugin_id, version, domain_id, upgrade_mode='AUTO'):
        """
        Returns: endpoint (message)
        """
        params = {
            'plugin_id': plugin_id,
            'version': version,
            'upgrade_mode': upgrade_mode,
            'domain_id': domain_id
        }

        meta: list = self.transaction.get_connection_meta()
        #meta.append(('token', self.system_key), )
        #_LOGGER.debug('[get_plugin_endpoint] meta:%s' % meta)
        response = self.client.Plugin.get_plugin_endpoint(
            params,
            metadata=meta
        )
        return response



import copy
import datetime
import logging
from spaceone.core.locator import Locator
from spaceone.core.scheduler import HourlyScheduler
from spaceone.inventory.service.cleanup_service import CleanupService
from spaceone.inventory.lib.scheduler import init_count, update_token, get_domain_id_from_token


__all__ = ['CleanupScheduler']

_LOGGER = logging.getLogger(__name__)


INTERVAL = 10


class CleanupScheduler(HourlyScheduler):

    def __init__(self, queue, interval, minute=':00'):
        super().__init__(queue, interval, minute)
        self.count = init_count()
        self.locator = Locator()
        self.TOKEN = update_token()
        self.domain_id = get_domain_id_from_token(self.TOKEN)

    def create_task(self):
        return [self._create_job_request(domain) for domain in self.list_domains()]

    def list_domains(self):
        try:
            if self.check_count() is False:
                # ERROR LOGGING
                pass

            # Loop all domain, then find scheduled collector
            metadata = {
                'token': self.TOKEN,
                'service': 'inventory',
                'resource': 'Cleanup',
                'verb': 'list_domains',
                'domain_id': self.domain_id
            }

            cleanup_svc: CleanupService = self.locator.get_service(CleanupService, metadata)
            response = cleanup_svc.list_domains({})
            return response.get('results', [])

        except Exception as e:
            _LOGGER.error(e)
            return []

    def check_count(self):
        current_time = datetime.datetime.utcnow()
        hour = current_time.hour

        if (self.count['hour'] + self.config) % 24 != hour:
            if self.count['hour'] == hour:
                _LOGGER.error('[check_count] duplicated call in the same time')
            else:
                _LOGGER.error('[check_count] missing time')

        self.count.update({
            'previous': current_time,
            'index': self.count['index'] + 1,
            'hour': hour,
            'started_at': current_time
        })

    def _update_count_ended_at(self):
        current_time = datetime.datetime.utcnow()
        self.count['ended_at'] = current_time

    def _create_job_request(self, domain):
        """ Based on domain, create Job Request
        Returns:
            jobs: SpaceONE Pipeline Template
        """
        # _LOGGER.debug(f'[_create_job_request] domain: {domain}')

        metadata = {
            'token': self.TOKEN,
            'service': 'inventory',
            'resource': 'Cleanup',
            'verb': 'update_job_state',
            'domain_id': self.domain_id
        }

        update_job_state = {
            'locator': 'SERVICE',
            'name': 'CleanupService',
            'metadata': copy.deepcopy(metadata),
            'method': 'update_job_state',
            'params': {
                'params': {
                    'options': {},
                    'domain_id': domain['domain_id']
                }
            }
        }

        metadata['verb'] = 'delete_resources'
        delete_resources = {
            'locator': 'SERVICE',
            'name': 'CleanupService',
            'metadata': copy.deepcopy(metadata),
            'method': 'delete_resources',
            'params': {
                'params': {
                    'options': {},
                    'domain_id': domain['domain_id']
                }
            }
        }

        metadata['verb'] = 'terminate_jobs'
        terminate_jobs = {
            'locator': 'SERVICE',
            'name': 'CleanupService',
            'metadata': copy.deepcopy(metadata),
            'method': 'terminate_jobs',
            'params': {
                'params': {
                    'options': {},
                    'domain_id': domain['domain_id']
                }
            }
        }

        metadata['verb'] = 'terminate_resources'
        terminate_resources = {
            'locator': 'SERVICE',
            'name': 'CleanupService',
            'metadata': copy.deepcopy(metadata),
            'method': 'terminate_resources',
            'params': {
                'params': {
                    'options': {
                    },
                    'domain_id': domain['domain_id']
                }
            }
        }

        stp = {
            'name': 'inventory_cleanup_schedule',
            'version': 'v1',
            'executionEngine': 'BaseWorker',
            'stages': [update_job_state, delete_resources, terminate_jobs, terminate_resources]
        }

        _LOGGER.debug(f'[_create_job_request] tasks: inventory_cleanup_schedule')
        return stp
